"""
Represents a :class:`Table` entity within the document.
Tables are hierarchical objects composed of :class:`TableCell` objects, which *implicitly* form columns and rows.

:class:`Table` object contains associated metadata within it. They include :class:`TableCell` information, headers, page number and 
page ID of the page within which it exists in the document. 
"""

import itertools
import logging
import os
import uuid
import xlsxwriter
from copy import deepcopy


from typing import List, Dict

from textractor.exceptions import InputError, MissingDependencyException
from textractor.entities.bbox import BoundingBox
from textractor.entities.table_cell import TableCell
from textractor.visualizers.entitylist import EntityList
from textractor.entities.document_entity import DocumentEntity
from textractor.entities.selection_element import SelectionElement
from textractor.entities.table_title import TableTitle
from textractor.entities.table_footer import TableFooter
from textractor.entities.word import Word
from textractor.utils.geometry_util import get_indices
from textractor.data.constants import SimilarityMetric, TextTypes, CellTypes, TableTypes
from textractor.data.constants import IS_COLUMN_HEAD, IS_MERGED_CELL
from textractor.utils.search_utils import SearchUtils, get_metadata_attr_name
from textractor.utils.text_utils import group_elements_horizontally, linearize_children
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.data.html_linearization_config import HTMLLinearizationConfig


class Table(DocumentEntity):
    """
    To create a new :class:`Table` object we need the following:

    :param entity_id:           Unique identifier of the table.
    :param bbox:                Bounding box of the table.
    """

    def __init__(self, entity_id, bbox: BoundingBox):
        super().__init__(entity_id, bbox)
        self.table_cells: List[TableCell] = []
        self._column_headers: Dict[str, List[TableCell]] = {}
        self._title: TableTitle = None
        self._footers: TableFooter = []
        self._table_type: TableTypes = TableTypes.UNKNOWN
        self._page = None
        self._page_id = None

    @property
    def words(self):
        """
        Returns all the :class:`Word` objects present in the :class:`Table`.

        :return words: List of Word objects, each representing a word within the Table.
        :rtype: EntityList[Word]
        """

        all_words = sum([cell.words for cell in self.table_cells], [])

        if not all_words:
            logging.info("Table contains no word entities.")

        return EntityList(all_words)

    @property
    def page(self):
        """
        :return: Returns the page number of the page the Table entity is present in.
        :rtype: int
        """

        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the Table entity.

        :param page_num: Page number where the Table entity exists.
        :type page_num: int
        """

        self._page = page_num

    @property
    def table_type(self):
        """
        :return: Returns the table type.
        :rtype: TableTypes
        """

        return self._table_type

    @table_type.setter
    def table_type(self, table_type: TableTypes):
        """
        Sets the table type attribute of the Table entity.

        :param title: Type of the Table entity.
        :type title: TableTypes
        """

        self._table_type = table_type

    @property
    def title(self):
        """
        :return: Returns the table title.
        :rtype: TableTitle
        """

        return self._title

    @title.setter
    def title(self, title: TableTitle):
        """
        Sets the table title attribute of the Table entity.

        :param title: Title of the Table entity.
        :type title: TableTitle
        """

        self._title = title

    @property
    def footers(self):
        """
        :return: Returns the table footers.
        :rtype: List[TableFooter]
        """

        return self._footers

    @footers.setter
    def footers(self, footers: List[TableFooter]):
        """
        Sets the footers attribute of the Table entity.

        :param footers: Footers of the Table entity.
        :type footers: List[TableFooter]
        """

        self._footers = footers

    @property
    def column_headers(self) -> Dict[str, List[TableCell]]:
        """
        :return: Returns the column headers of the Table entity.
        :rtype: Dict[str, List[TableCell]]
        """

        return self._column_headers

    @column_headers.setter
    def column_headers(self, column_headers: Dict[str, List[TableCell]]):
        """
        Sets the column headers of the Table entity

        :param column_headers: Column headers
        :type column_headers: Dict[str, List[TableCell]]
        """

        self._column_headers = column_headers

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """

        return self._page_id

    @property
    def checkboxes(self) -> List[SelectionElement]:
        checkboxes = []
        for cell in self.table_cells:
            checkboxes.extend(cell.checkboxes)
        return checkboxes

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the Table entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """

        self._page_id = page_id

    @property
    def column_count(self):
        return max(list(self._get_table_cells(row_wise=False, column_wise=True).keys()))

    @property
    def row_count(self):
        return max(list(self._get_table_cells().keys()))

    def get_words_by_type(self, text_type=TextTypes.PRINTED):
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes

        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        if not self.words:
            return []

        return EntityList([word for word in self.words if word.text_type == text_type])

    def get_table_range(self):
        """
        :return: Returns the number of rows and columns in the table.
        :rtype: Tuple(int)
        """

        return self.row_count, self.column_count

    def strip_headers(
        self,
        column_headers: bool = True,
        in_table_title: bool = False,
        section_titles=False,
    ):
        """
        Returns a new :class:`Table` object after removing all cells that are marked as column headers in the table from the API response.

        :param column_headers: Remove the column headers
        :type column_headers: bool
        :param in_table_title: Remove the in-table titles
        :type in_table_title: bool
        :param section_titles: Remove the in-table section titles
        :type section_titles: bool

        :return: Table object after removing the headers.
        :rtype: Table
        """

        table_cells = deepcopy(self.table_cells)
        table_cells = [
            cell
            for cell in table_cells
            if not (
                (cell.is_column_header and column_headers)
                or (cell.is_title and in_table_title)
                or (cell.is_section_title and section_titles)
            )
        ]

        old_rows = set([cell.row_index for cell in table_cells])
        new_rows = set(range(1, len(old_rows) + 1))

        new_row_mapping = dict(zip(old_rows, new_rows))

        new_table_cells = []
        for cell in table_cells:
            cell.row_index = new_row_mapping[cell.row_index]
            new_table_cells.append(cell)

        new_table = deepcopy(self)
        new_table.add_cells(new_table_cells)

        return new_table

    def add_cells(self, cells: List[TableCell]):
        """
        Add :class:`TableCell` objects to the :class:`Table`.
        This function does not check the integrity of the table after the cells are added.

        :param cells: List of TableCell objects, each representing a single cell within the table.
                      No specific ordering is assumed since it is implicitly ordered by row and column index.
        :type cells: list
        """

        for cell in cells:
            cell.table_id = self.id
        self.table_cells = cells
        self.table_cells = sorted(
            self.table_cells, key=lambda x: (x.row_index, x.col_index)
        )

    def _get_table_cells(
        self, row_wise: bool = True, column_wise: bool = False
    ) -> Dict[int, List[TableCell]]:
        """
        Return a dictionary of row_number/column_number : List[TableCell] (in order) for the table entity. Only one of the 2 params should be set to True. default: row_wise=True, column_wise=False.

        :param row_wise: True if format of dict `{row_number : List[TableCell]}`. Else False.
        :type row_wise: bool
        :param column_wise: True if format of dict `{row_number : List[TableCell]}`. Else False.
        :type column_wise: bool

        :return: Returns a dictionary of format `{row_number/column_number : List[TableCell]}`
        :rtype: Dict[int, List[TableCell]]
        """

        if (row_wise and column_wise) or (not row_wise and not column_wise):
            logging.warning(
                "Both parameters cannot be equal. Choose one (row_wise=True or column_wise=True)"
            )
            return {}

        attribute = "row_index" if row_wise else "col_index"
        alt_attribute = "col_index" if row_wise else "row_index"
        sorted_cells = {}
        for cell in self.table_cells:
            index = getattr(cell, attribute)
            if index in sorted_cells.keys():
                sorted_cells[index].append(cell)
            else:
                sorted_cells[index] = [cell]

        for key in sorted_cells.keys():
            sorted_cells[key] = sorted(
                sorted_cells[key], key=lambda x: getattr(x, alt_attribute)
            )

        return sorted_cells

    def get_cells_by_type(self, cell_type: CellTypes = CellTypes.COLUMN_HEADER):
        """
        Returns a dictionary of column_header (str) : List[TableCell] (in order).

        :param cell_type: supports CellTypes.COLUMN_HEADER as of now, will support SECTION_TITLE, FLOATING_TITLE,
                            FLOATING_FOOTER, SUMMARY_CELL in the future.
        :type cell_type: CellTypes

        :return: {column_header (str) : List[TableCell]}
        :rtype: Dict[str, List[TableCell]]
        """
        if not isinstance(cell_type, CellTypes):
            raise InputError(
                "cell_type parameter should be of CellTypes type. Find input choices from textractor.data.constants"
            )

        cell_property = get_metadata_attr_name(cell_type)
        if cell_property == "":
            logging.info("No cells of this type exist.")
            return []
        filtered_cells = {}

        for cell in self.table_cells:
            if cell.metadata[cell_property]:
                if not cell.metadata[IS_MERGED_CELL]:
                    header = " ".join([word.text for word in cell.words])
                    filtered_cells[header] = [cell]
                else:
                    all_merge_cells = sorted(
                        cell.siblings, key=lambda x: x.bbox.x + x.bbox.y
                    )
                    all_words = sum([cell.words for cell in all_merge_cells], [])
                    header = " ".join([word.text for word in all_words])
                    filtered_cells[header] = all_merge_cells

        if cell_type == CellTypes.COLUMN_HEADER and not self._column_headers:
            self._column_headers = filtered_cells

            if not self._column_headers:
                logging.info("Column headers have not been identified in this table.")

        return filtered_cells

    def get_columns_by_name(
        self,
        column_names,
        similarity_metric=SimilarityMetric.COSINE,
        similarity_threshold=0.6,
    ):
        """
        Returns a dictionary of format {column_name : List[TableCell]} for the column names listed in param column_names.

        :param column_names: List of column names of columns to be extracted from table.
        :type column_names: list
        :param similarity_metric: 'cosine', 'euclidean' or 'levenshtein'. 'cosine' is chosen as default.
        :type similarity_metric: str
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a new Table consisting of columns passed in column_names.
        :rtype: Table
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )

        table_columns = (
            self._column_headers.keys()
            if self._column_headers
            else self.get_cells_by_type().keys()
        )
        column_indices = set()
        chosen_columns = []
        similarity_threshold = (
            similarity_threshold
            if similarity_metric == SimilarityMetric.COSINE
            else -(similarity_threshold)
        )
        for column in column_names:
            for table_column in table_columns:
                similarity = SearchUtils.get_word_similarity(
                    column, table_column, similarity_metric
                )
                similarity = (
                    similarity
                    if similarity_metric == SimilarityMetric.COSINE
                    else -(similarity)
                )

                if similarity > similarity_threshold:
                    chosen_columns.append(table_column)

        for col in chosen_columns:
            for cell in self._column_headers[col]:
                column_indices.add(cell.col_index)

        table_cells = deepcopy(self.table_cells)
        table_cells = [cell for cell in table_cells if cell.col_index in column_indices]
        new_cols = set(range(1, len(column_indices) + 1))
        new_col_mapping = dict(zip(column_indices, new_cols))

        new_table_cells = []
        for cell in table_cells:
            cell.col_index = new_col_mapping[cell.col_index]
            new_table_cells.append(cell)

        new_table = deepcopy(self)
        new_table.add_cells(new_table_cells)

        return new_table

    def __getitem__(self, key):
        """
        Returns a new :class:`Table` with selected rows and columns. The table can be filtered using numpy indexing.

        :param key: Tuple[rows, columns] with numpy indexing
                    Eg: table[0:3, :] returns all columns of the first 3 rows of the Table.

        :return: Returns a new Table with selected rows and columns.
        :rtype: Table
        """

        np_rows, np_cols = key

        if np_rows == np_cols and np_rows == slice(None, None, None):
            return self

        max_rows, max_cols = self.get_table_range()

        row_index = (
            str(np_rows)
            if isinstance(np_rows, int)
            else ":".join([str(np_rows.start), str(np_rows.stop), str(np_rows.step)])
        )
        col_index = (
            str(np_cols)
            if isinstance(np_cols, int)
            else ":".join([str(np_cols.start), str(np_cols.stop), str(np_cols.step)])
        )

        rows = get_indices(row_index, max_rows)
        cols = get_indices(col_index, max_cols)

        table_rows = self._get_table_cells(row_wise=True, column_wise=False)

        filtered_rows = {}
        for row_number in rows:
            if row_number + 1 in table_rows.keys():
                row_data = table_rows[row_number + 1]
                col_cells = []
                for col_number in cols:
                    col_cells.append(row_data[col_number])
                filtered_rows[row_number] = col_cells

        new_table_cells = _get_new_table_cells(rows, filtered_rows)

        new_table = deepcopy(self)
        new_table.add_cells(new_table_cells)

        return new_table

    def to_pandas(self, use_columns=False, config: TextLinearizationConfig = TextLinearizationConfig()):
        """
        Converts the table to a pandas DataFrame

        :param use_columns: If the first row of the table is made of column headers, use them for the pandas dataframe. Only supports single row header.
        :param config: Text linearization configuration object for the table content
        :return:
        """
        try:
            from pandas import DataFrame
        except ImportError:
            raise MissingDependencyException(
                "pandas library is required for exporting tables to DataFrame objects or markdown"
            )

        rows = sorted([(key, list(group)) for key, group in itertools.groupby(
            self.table_cells, key=lambda cell: cell.row_index
        )], key=lambda r: r[0])
        row_offset = 0

        columns = None
        processed_cells = set()
        if use_columns:
            # Try to automatically get the columns if they are in the first row
            columns = [[] for _ in range(self.column_count)]
            is_header_count = 0
            for _, row in rows:
                if not any([c.is_column_header for c in row]):
                    # There is not header in that row, we are done
                    break
                for i, cell in enumerate(row):
                    if (
                        cell not in processed_cells or
                        config.table_duplicate_text_in_merged_cells or
                        config.table_flatten_headers
                    ):
                        if cell.siblings:
                            # This handles the edge case where we are flattening the headers
                            # so we want to duplicate the cell text but only in its first row
                            first_row, _, _, _ = cell._get_merged_cell_range()
                            if cell in processed_cells and first_row != cell.row_index:
                                continue
                            children = []
                            for sib in cell.siblings:
                                if sib.is_column_header:
                                    is_header_count += 1
                                children.extend(sib.children)
                                processed_cells.add(sib)
                            text, _ = linearize_children(children, config=config, no_new_lines=True)
                            columns[i].append(text)
                        else:
                            if cell.is_column_header:
                                is_header_count += 1
                            text = cell.get_text(config)
                            columns[i].append(text)
                    elif config.table_cell_empty_cell_placeholder:
                        columns[i].append(config.table_cell_empty_cell_placeholder)
                    else:
                        columns[i].append("")
                row_offset += 1
            # If we have the correct number of column and at least half the row is tagged as a header
            if len(columns) == self.column_count and is_header_count / len(columns) >= config.table_column_header_threshold:
                use_columns = True
            else:
                use_columns = False
                logging.info(
                    f"The number of column header cell do not match the column count, ignoring them, {len(columns)} vs {self.column_count}"
                )

        if columns and any([c for c in columns]) and config.table_flatten_headers:
            columns = ["".join(c) for c in columns]
            table = [columns]
        elif columns and any([c for c in columns]):
            # We reset the row offset as only the first line will be taken as header
            columns = [c[0] for c in columns]
            table = [columns]
            row_offset = 1
        else:
            table = []

        for _, row in rows[row_offset:]:
            table.append([])
            for cell in row:
                table[-1].append("")
                if cell.siblings:
                    children = []
                    first_row, first_col, last_row, last_col = cell._get_merged_cell_range()
                    if (cell.col_index == first_col and cell.row_index == first_row) or config.table_duplicate_text_in_merged_cells:
                        for sib in cell.siblings:
                            children.extend(sib.children)
                            processed_cells.add(sib)
                        text, _ = linearize_children(children, config=config, no_new_lines=True)
                    elif cell.row_index == first_row and config.table_cell_left_merge_cell_placeholder:
                        text = config.table_cell_left_merge_cell_placeholder
                    elif cell.col_index == first_col and config.table_cell_top_merge_cell_placeholder:
                        text = config.table_cell_top_merge_cell_placeholder
                    elif cell.col_index != first_col and cell.row_index != first_row and config.table_cell_cross_merge_cell_placeholder:
                        text = config.table_cell_cross_merge_cell_placeholder
                    else:
                        text = config.table_cell_empty_cell_placeholder if config.table_cell_empty_cell_placeholder else ""
                else:
                    text = cell.get_text(config)
                table[-1][cell.col_index - 1] = text if text or not config.table_cell_empty_cell_placeholder else config.table_cell_empty_cell_placeholder

        return DataFrame(
            table[1:] if use_columns else table,
            columns=columns if use_columns else None,
        )

    def to_csv(self, use_columns = False, config: TextLinearizationConfig = TextLinearizationConfig()) -> str:
        """Returns the table in the Comma-Separated-Value (CSV) format

        :param use_columns: If the first row of the table is made of column headers, use them for the pandas dataframe. Only supports single row header.
        :param config: Text linearization configuration object for the table content
        :return: Table as a CSV string.
        :rtype: str
        """
        return self.to_pandas(use_columns=use_columns, config=config).to_csv()

    def to_excel(self, filepath=None, workbook=None, save_workbook=True):
        """
        Export the Table Entity as an excel document. Advantage of excel over csv is that it can accommodate merged cells that we see so often with Textract documents.

        :param filepath: Path to store the exported Excel file
        :type filepath: str
        :param workbook: if xlsxwriter workbook is passed to the function, the table is appended to the last sheet of that workbook.
        :type workbook: xlsxwriter.Workbook
        :param save_workbook: Flag to save_notebook. If False, it is returned by the function.
        :type save_workbook: bool

        :return: Returns a workbook if save_workbook is False. Else, saves the .xlsx file in the filepath if was initialized with.
        :rtype: xlsxwriter.Workbook
        """
        if workbook is None:
            if filepath is not None:
                workbook = xlsxwriter.Workbook(filepath)
            else:
                logging.error("You need to provide a filepath or a workbook")

        worksheet = workbook.add_worksheet()

        merged_cells = set()
        for cell in self.table_cells:
            cell_content = cell.__repr__().split(">")[1][1:]
            if cell.metadata[IS_MERGED_CELL]:
                first_row, first_col, last_row, last_col = cell._get_merged_cell_range()
                if (first_row, first_col, last_row, last_col) in merged_cells:
                    continue
                else:
                    merged_cells.add((first_row, first_col, last_row, last_col))
                worksheet.merge_range(
                    first_row - 1,
                    first_col - 1,
                    last_row - 1,
                    last_col - 1,
                    cell_content,
                )
            else:
                worksheet.write_string(
                    cell.row_index - 1, cell.col_index - 1, cell_content
                )

        if save_workbook:
            workbook.close()

        else:
            return workbook

    def to_html(self) -> str:
        """Returns the table in the HTML format

        :return: Table as an HTML string.
        :rtype: str
        """
        
        return self.get_text(HTMLLinearizationConfig())

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        local_config = deepcopy(config)
        words_ = self.words
        # If no text, return empty string
        if not words_ and local_config.table_remove_column_headers:
            return "", []

        # If not many words, only return text
        if len(words_) < local_config.table_min_table_words:
            return linearize_children(words_, config=config)

        words = [Word(str(uuid.uuid4()), self.bbox, local_config.table_prefix)] if local_config.table_prefix else []
        rows = sorted([(key, list(group)) for key, group in itertools.groupby(
            self.table_cells, key=lambda cell: cell.row_index
        )], key=lambda r: r[0])
        processed_cells = set()
        # Fill the table
        row_offset = 0
        if local_config.table_flatten_headers:
            columns = [[] for _ in range(len(rows[0][1]))]
            columns_bbox = [[] for _ in range(len(rows[0][1]))]
            for _, row in rows:
                if not any([c.is_column_header for c in row]):
                    # There is not header in that row, we are done
                    break
                for i, cell in enumerate(row):
                    if (
                        cell not in processed_cells or
                        local_config.table_duplicate_text_in_merged_cells or
                        local_config.table_flatten_headers
                    ):
                        if cell.siblings:
                            # This handles the edge case where we are flattening the headers
                            # so we want to duplicate the cell text but only in its first row
                            first_row, _, _, _ = cell._get_merged_cell_range()
                            if cell in processed_cells and first_row != cell.row_index:
                                continue
                            children = []
                            for sib in cell.siblings:
                                children.extend(sib.children)
                                processed_cells.add(sib)
                            _, words = linearize_children(children, config=config, no_new_lines=True)
                            columns[i].extend(words)
                            columns_bbox[i].append(cell.bbox)
                        else:
                            _, words = cell.get_text_and_words(config)
                            columns[i].extend(words)
                            columns_bbox[i].append(cell.bbox)
                    elif local_config.table_cell_empty_cell_placeholder:
                        columns[i].append(Word(str(uuid.uuid4()), cell.bbox, local_config.table_cell_empty_cell_placeholder))
                row_offset += 1
            if columns:
                columns_bbox = [BoundingBox.enclosing_bbox(cbb) for cbb in columns_bbox]
                if local_config.table_row_prefix and local_config.add_prefixes_and_suffixes_as_words:
                    words.append(Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(columns_bbox), local_config.table_row_prefix, is_structure=True))
                for i, column in enumerate(columns):
                    words.append(
                        Word(
                            str(uuid.uuid4()),
                            columns_bbox[i],
                            local_config.table_cell_header_prefix
                            if local_config.table_cell_header_prefix
                            else local_config.table_cell_prefix,
                            is_structure=True
                        )
                    )
                    words.extend(column)
                    words.append(
                        Word(
                            str(uuid.uuid4()),
                            columns_bbox[i],
                            local_config.table_cell_header_suffix
                            if local_config.table_cell_header_suffix
                            else local_config.table_cell_suffix,
                            is_structure=True
                        )
                    )
                if local_config.table_row_suffix and local_config.add_prefixes_and_suffixes_as_words:
                    words.append(Word(str(uuid.uuid4()), columns_bbox, local_config.table_row_suffix, is_structure=True))
        for _, cells in rows[row_offset:]:
            if local_config.table_row_prefix and local_config.add_prefixes_and_suffixes_as_words:
                words.append(Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(cells), local_config.table_row_prefix, is_structure=True))
            for cell in sorted(cells, key=lambda c: c.col_index):
                # Siblings includes the current cell
                if cell.siblings:
                    first_row, first_col, last_row, last_col = cell._get_merged_cell_range()
                    cell_id = cell.id
                    cell_bbox = BoundingBox.enclosing_bbox(cell.siblings)
                    col_index = first_col
                    col_span = last_col - first_col + 1
                    row_index = first_row
                    row_span = last_row - first_row + 1
                    children = []
                    if (cell.col_index == first_col and cell.row_index == first_row) or local_config.table_duplicate_text_in_merged_cells:
                        for sib in cell.siblings:
                            children.extend(sib.children)
                            processed_cells.add(sib)
                        _, cell_words = linearize_children(children, config=config, no_new_lines=True)
                    elif cell.row_index == first_row and local_config.table_cell_left_merge_cell_placeholder:
                        # Left-merge token
                        cell_words = [
                            Word(str(uuid.uuid4()),
                                cell_bbox,
                                local_config.table_cell_left_merge_cell_placeholder,
                                is_structure=True
                            )
                        ]
                    elif cell.col_index == first_col and local_config.table_cell_top_merge_cell_placeholder:
                        # Top-merge token
                        cell_words = [
                            Word(str(uuid.uuid4()),
                                cell_bbox,
                                local_config.table_cell_top_merge_cell_placeholder,
                                is_structure=True
                            )
                        ]
                    elif cell.col_index != first_col and cell.row_index != first_row and local_config.table_cell_cross_merge_cell_placeholder:
                        # Cross-merge token (left and top)
                        cell_words = [
                            Word(str(uuid.uuid4()),
                                cell_bbox,
                                local_config.table_cell_cross_merge_cell_placeholder,
                                is_structure=True
                            )
                        ]
                    else:
                        continue
                else:
                    cell_id = cell.id
                    cell_bbox = cell.bbox
                    col_index = cell.col_index
                    col_span = cell.col_span
                    row_index = cell.row_index
                    row_span = cell.row_span
                    _, cell_words = cell.get_text_and_words(config)
                if local_config.add_prefixes_and_suffixes_as_words:
                    if local_config.table_cell_prefix or (local_config.table_cell_header_prefix and cell.is_column_header):
                        words.append(
                            Word(
                                str(uuid.uuid4()),
                                cell_bbox,
                                local_config.table_cell_header_prefix
                                if cell.is_column_header and local_config.table_cell_header_prefix
                                else local_config.table_cell_prefix,
                                is_structure=True
                            )
                        )
                        words[-1].cell_id = cell_id
                        words[-1].cell_bbox = cell_bbox
                        words[-1].col_index = col_index
                        words[-1].col_span = col_span
                        words[-1].row_index = row_index
                        words[-1].row_span = row_span

                    words.extend(cell_words)
                    if not cell_words and local_config.table_cell_empty_cell_placeholder:
                        words.append(Word(str(uuid.uuid4()), cell_bbox, local_config.table_cell_empty_cell_placeholder))

                    if local_config.table_cell_suffix or (local_config.table_cell_header_suffix and cell.is_column_header):
                        words.append(
                            Word(
                                str(uuid.uuid4()),
                                cell_bbox,
                                local_config.table_cell_header_suffix if cell.is_column_header and local_config.table_cell_header_suffix else local_config.table_cell_suffix,
                                is_structure=True
                            )
                        )
                        words[-1].cell_id = cell_id
                        words[-1].cell_bbox = cell_bbox
                        words[-1].col_index = col_index
                        words[-1].col_span = col_span
                        words[-1].row_index = row_index
                        words[-1].row_span = row_span
                else:
                    words.extend(cell_words)
            if local_config.table_row_suffix and local_config.add_prefixes_and_suffixes_as_words:
                words.append(Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(cells), local_config.table_row_suffix, is_structure=True))

        if local_config.table_suffix:
            words.append(Word(str(uuid.uuid4()), self.bbox, local_config.table_suffix))

        for w in words:
            w.table_id = str(self.id)
            w.table_bbox = self.bbox

        text = (local_config.table_prefix if local_config.add_prefixes_and_suffixes_in_text else "")
        # Markdown
        if local_config.table_linearization_format == "markdown":
            df = self.to_pandas(
                use_columns=True,
                config=config
            )
            has_column = any([isinstance(c, str) for c in df.columns])
            if local_config.table_remove_column_headers:
                headers = df.columns if has_column else ["" for c in df.columns]
            else:
                headers = df.columns
            table = df.to_markdown(
                tablefmt=local_config.table_tabulate_format, headers=headers, index=False
            )
            if local_config.table_tabulate_remove_extra_hyphens:
                while "-" * 2 in table:
                    table = table.replace("--", "-")
            text += table
        # Plaintext or HTML
        else:
            # FIXME: The cyclomatic complexity of doing things like this will be unsustainable.
            if local_config.table_flatten_semi_structured_as_plaintext and self.table_type == TableTypes.SEMI_STRUCTURED:
                text = "<p>"
                if config.table_linearization_format == "html":
                    local_config.table_prefix = "<p>"
                    local_config.table_suffix = "</p>"
                else:
                    local_config.table_prefix = ""
                    local_config.table_suffix = ""
                local_config.table_linearization_format = "plaintext"
                local_config.table_row_prefix = ""
                local_config.table_row_suffix = ""
                local_config.table_cell_header_prefix = ""
                local_config.table_cell_header_suffix = ""
                local_config.table_cell_prefix = ""
                local_config.table_cell_suffix = " "
            elif local_config.table_linearization_format == "html":
                local_config.table_prefix = "<table>"
                local_config.table_suffix = "</table>"
                local_config.table_row_prefix = "<tr>"
                local_config.table_row_suffix = "</tr>"
                local_config.table_cell_header_prefix = "<th>"
                local_config.table_cell_header_suffix = "</th>"
                local_config.table_cell_prefix = "<td>"
                local_config.table_cell_suffix = "</td>"
                
            row_offset = 0
            processed_cells = set()
            if local_config.table_flatten_headers:
                columns = ["" for _ in range(len(rows[0][1]))]
                column_spans = [1 for _ in range(len(rows[0][1]))]
                for _, row in rows:
                    if not any([c.is_column_header for c in row]):
                        # There is not header in that row, we are done
                        break
                    for i, cell in enumerate(row):
                        if (
                            cell not in processed_cells or
                            local_config.table_duplicate_text_in_merged_cells or
                            local_config.table_flatten_headers
                        ):
                            if cell.siblings:
                                # This handles the edge case where we are flattening the headers
                                # so we want to duplicate the cell text but only in its first row
                                first_row, first_col, _, last_col = cell._get_merged_cell_range()
                                # We always compute the colspan, but we will discard them if config.table_duplicate_text_in_merged_cells is True
                                column_spans[i] = last_col - first_col + 1
                                if cell in processed_cells and first_row != cell.row_index:
                                    continue
                                children = []
                                for sib in cell.siblings:
                                    children.extend(sib.children)
                                    processed_cells.add(sib)
                                text, _ = linearize_children(children, config=local_config, no_new_lines=True)
                                columns[i] += text
                            else:
                                text = cell.get_text(local_config)
                                columns[i] += text
                        elif local_config.table_cell_empty_cell_placeholder and local_config.table_linearization_format != "html":
                            columns[i] += local_config.table_cell_empty_cell_placeholder
                        else:
                            columns[i] += ""
                            column_spans[i] = 0
                    row_offset += 1
                if any(columns):
                    text += local_config.table_row_prefix if local_config.add_prefixes_and_suffixes_in_text else ""
                    for column, column_span in zip(columns, column_spans):
                        if column_span == 0:
                            continue
                        
                        if local_config.table_linearization_format == "html":
                            prefix = f'<th colspan="{column_span}">'
                        else:
                            prefix = (local_config.table_cell_header_prefix if local_config.table_cell_header_prefix else local_config.table_cell_prefix)
                        # We don't have any rowspan logic here as the flattened header will have rowspan=1 always.
                        text += (
                            (
                                prefix + 
                                (column or local_config.table_cell_empty_cell_placeholder) +
                                (local_config.table_cell_header_suffix if local_config.table_cell_header_suffix else local_config.table_cell_suffix)
                            )
                            if local_config.add_prefixes_and_suffixes_in_text or local_config.table_linearization_format == "html" else
                            (column or local_config.table_cell_empty_cell_placeholder)
                        )
                        text += local_config.table_column_separator
                    text += (local_config.table_row_suffix if local_config.add_prefixes_and_suffixes_in_text else "")
                    text += local_config.table_row_separator
                    
            for _, row in rows[row_offset:]:
                text += (local_config.table_row_prefix if local_config.add_prefixes_and_suffixes_in_text else "")
                for cell in sorted(row, key=lambda c: c.col_index):
                    # This will return row_index, col_index, row_index, col_index for regular cells
                    first_row, first_col, last_row, last_col = cell._get_merged_cell_range()
                    if cell in processed_cells and local_config.table_linearization_format == "html":
                        continue
                    
                    # Siblings includes the current cell
                    if cell.siblings:
                        children = []
                        if (cell.col_index == first_col and cell.row_index == first_row) or local_config.table_duplicate_text_in_merged_cells:
                            for sib in cell.siblings:
                                children.extend(sib.children)
                                processed_cells.add(sib)
                            cell_text, _ = linearize_children(children, config=local_config, no_new_lines=True)
                        elif cell.row_index == first_row and local_config.table_cell_left_merge_cell_placeholder:
                            cell_text = local_config.table_cell_left_merge_cell_placeholder
                        elif cell.col_index == first_col and local_config.table_cell_top_merge_cell_placeholder:
                            cell_text = local_config.table_cell_top_merge_cell_placeholder
                        elif cell.col_index != first_col and cell.row_index != first_row and local_config.table_cell_cross_merge_cell_placeholder:
                            cell_text = local_config.table_cell_cross_merge_cell_placeholder
                        else:
                            cell_text = ""
                            
                    # This is a regular cell
                    else:
                        cell_text, _ = cell.get_text_and_words(local_config)
                    
                    if local_config.table_linearization_format == "html":
                        if cell.is_column_header:
                            prefix = (
                                '<th' +
                                (f' colspan="{last_col - first_col + 1}" ' if last_col - first_col > 0 else '') +
                                (f' rowspan="{last_row - first_row + 1}"' if last_row - first_row > 0 else '') +
                                '>'
                            )
                        else:
                            prefix = (
                                '<td' +
                                (f' colspan="{last_col - first_col + 1}" ' if last_col - first_col > 0 else '') +
                                (f' rowspan="{last_row - first_row + 1}"' if last_row - first_row > 0 else '') +
                                '>'
                            )
                    else:
                        prefix = local_config.table_cell_header_prefix if local_config.table_cell_header_prefix and cell.is_column_header else local_config.table_cell_prefix
                        
                    text += (
                        (
                            prefix +
                            # Removes trailing whitespace in cell_text
                            (cell_text.strip() or local_config.table_cell_empty_cell_placeholder) +
                            (local_config.table_cell_header_suffix if local_config.table_cell_header_suffix and cell.is_column_header else local_config.table_cell_suffix)
                        )
                        if local_config.add_prefixes_and_suffixes_in_text else
                        (cell_text or local_config.table_cell_empty_cell_placeholder)
                    ) 
                    text += local_config.table_column_separator
                if text and text[-1] == local_config.table_column_separator:
                    text = text[:-1]
                text += (local_config.table_row_suffix if local_config.add_prefixes_and_suffixes_in_text else "")
                text += local_config.table_row_separator
                
        if local_config.table_add_title_as_caption and self.title and local_config.table_linearization_format == "html":
            text += "<caption>" + self.title.get_text() + "</caption>"
            
        text += (local_config.table_suffix if local_config.add_prefixes_and_suffixes_in_text else "")
        
        return text, words

    def to_txt(self):
        table = [["" for _ in range(self.column_count)] for _ in range(self.row_count)]
        for cell in self.table_cells:
            table[cell.row_index - 1][cell.col_index - 1] = cell.text

        return os.linesep.join(["\t".join(r) for r in table])

    def __repr__(self):
        return os.linesep.join(
            [
                "Table",
                f"Rows - {self.row_count}",
                f"Columns - {self.column_count}",
                f"Cells - {len(self.table_cells)}",
                f"Merged Cells - {len([c for c in self.table_cells if c.metadata[IS_MERGED_CELL]])}",
            ]
        )


def _get_new_table_cells(rows, filtered_rows):
    """
    Modifies indexing of :class:`TableCell` entities according to the new filtered table format.

    :param rows: List containing the filtered row indices from original table.
    :type rows: list
    :param filtered_rows: Dictionary containing filtered rows and respective cells. Format row : List[TableCell]
    :type filtered_rows: dict

    :return: List of modified TableCell objects with new indexing.
    :rtype: List
    """

    rows = sorted([r + 1 for r in rows])
    cols = sorted(list(set([cell.col_index for cell in filtered_rows[rows[0] - 1]])))

    new_row_range = list(range(1, len(rows) + 1))
    new_col_range = list(range(1, len(filtered_rows[rows[0] - 1]) + 1))

    row_index_map = dict(zip(rows, new_row_range))
    col_index_map = dict(zip(cols, new_col_range))

    all_cells = sum(list(filtered_rows.values()), [])
    old_cells_dict = {(cell.row_index, cell.col_index): cell for cell in all_cells}

    new_cells_dict = {
        (row_index_map[cell.row_index], col_index_map[cell.col_index]): TableCell(
            entity_id=cell.id,
            bbox=cell.bbox,
            row_index=row_index_map[cell.row_index],
            col_index=col_index_map[cell.col_index],
            row_span=1,
            col_span=1,
            confidence=cell.confidence,
        )
        for cell in old_cells_dict.values()
    }

    for cell in old_cells_dict.values():
        new_cell_siblings = []
        cell_words = []
        cell_children = []

        cell_row, cell_col = row_index_map.get(cell.row_index), col_index_map.get(
            cell.col_index
        )
        new_cells_dict[(cell_row, cell_col)].metadata = cell.metadata

        if (cell_row, cell_col) in new_cells_dict.keys() and not cell.metadata[
            IS_MERGED_CELL
        ]:
            new_cells_dict[(cell_row, cell_col)].add_children(list(cell.children))

        else:
            if (cell_row, cell_col) in new_cells_dict.keys() and not new_cells_dict[
                (cell_row, cell_col)
            ].siblings:
                for subcell in cell.siblings:
                    cell_words.extend(list(subcell.words))
                    cell_children.extend(list(subcell.children))

                    new_row, new_col = row_index_map.get(
                        subcell.row_index
                    ), col_index_map.get(subcell.col_index)

                    if (new_row, new_col) in new_cells_dict.keys():
                        new_cells_dict[(new_row, new_col)].metadata = subcell.metadata
                        new_cell_siblings.append(new_cells_dict[(new_row, new_col)])

                new_cells_dict[(cell_row, cell_col)].siblings = new_cell_siblings
                new_cells_dict[(cell_row, cell_col)].add_children(cell_children)

                for sibling in new_cells_dict[(cell_row, cell_col)].siblings:
                    sibling.siblings = new_cell_siblings
            else:
                continue

    return list(new_cells_dict.values())



"""
:class:`Linearizable` is a class that defines how a component can be linearized (converted to text)
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.data.html_linearization_config import HTMLLinearizationConfig
from textractor.data.markdown_linearization_config import MarkdownLinearizationConfig

class Linearizable(ABC):    
    def get_text(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> str:
        """
        Returns the linearized text of the entity

        :param config: Text linearization confi 
        :type config:   
        :return: Linearized text of the entity
        :rtype: str
        """
        text, _ = self.get_text_and_words(config=config)
        return text

    @property
    def text(self) -> str:
        """
        Maps to .get_text()

        :return: Returns the linearized text of the entity
        :rtype: str
        """
        return self.get_text()

    def to_html(
        self,
        config: HTMLLinearizationConfig = HTMLLinearizationConfig()
    ) -> str:
        """
        Returns the HTML representation of the entity

        :return: HTML text of the entity
        :rtype: str
        """
        return self.get_text(config)

    def to_markdown(
        self,
        config: MarkdownLinearizationConfig = MarkdownLinearizationConfig()
    ) -> str:
        """
        Returns the markdown representation of the entity

        :return: Markdown text of the entity
        :rtype: str
        """
        return self.get_text(config)

    @abstractmethod
    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the entity and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        pass



"""The Document class is defined to host all the various DocumentEntity objects within it. :class:`DocumentEntity` objects can be 
accessed, searched and exported the functions given below."""

import boto3
import time
from typing import Any

from textractor.entities.document import Document
from textractor.parsers.response_parser import parse
from textractor.data.constants import TextractAPI
from textractor.utils.results_utils import results_exist, get_full_json_from_output_config
from textractcaller.t_call import (
    get_job_response,
    get_full_json,
    OutputConfig,
)


class LazyDocument:
    """
    LazyDocument is a proxy for Document when using the async APIs. It will not load the response until
    one if its property is used. You can access the underlying Document object using the document property.
    """

    def __init__(
        self,
        job_id: str,
        api: TextractAPI,
        textract_client=None,
        images=None,
        output_config: OutputConfig = None,
    ):
        """
        Creates a new document, ideally containing entity objects pertaining to each page.

        :param num_pages: Number of pages in the input Document.
        """
        self.job_id = job_id
        self._api = api
        self._textract_client = textract_client
        self._document = None
        self._images = images
        self._output_config = output_config
        self._s3_polling_interval = 1
        self._textract_polling_interval = 5

    @property
    def s3_polling_interval(self) -> int:
        """Getter for the polling interval

        :return: Time between get_full_result calls
        :rtype: int
        """
        return self._s3_polling_interval

    @s3_polling_interval.setter
    def s3_polling_interval(self, s3_polling_interval: int):
        """Setter for the polling interval

        :param polling_interval: Time between get_full_result calls
        :type polling_interval: int
        """
        self._s3_polling_interval = s3_polling_interval

    @property
    def textract_polling_interval(self) -> int:
        """Getter for the polling interval

        :return: Time between get_full_result calls
        :rtype: int
        """
        return self._textract_polling_interval

    @textract_polling_interval.setter
    def textract_polling_interval(self, textract_polling_interval: int):
        """Setter for the polling interval

        :param polling_interval: Time between get_full_result calls
        :type polling_interval: int
        """
        self._textract_polling_interval = textract_polling_interval

    @property
    def document(self) -> Document:
        """Getter for the underlying Document object

        :return: Proxied Document object
        :rtype: Document
        """
        return object.__getattribute__(self, "_document")

    def __getattr__(self, __name: str) -> Any:
        """Proxy that gets the Document result when a document property is
        accessed. This a blocking call.

        :param __name: Name of the attribute
        :type __name: str
        :return: Value of the attribute
        :rtype: Any
        """

        # Prevents infinite recursion on LazyDocument properties
        if __name in [
            "job_id",
            "_api",
            "_textract_client",
            "_document",
            "_images",
            "s3_polling_interval",
            "_s3_polling_interval",
            "textract_polling_interval",
            "_textract_polling_interval",
            "_output_config",
        ]:
            return object.__getattribute__(self, __name)

        if self._document is None:
            if self._output_config:
                s3_client = boto3.client("s3")
                start = time.time()
                response = None
                while not results_exist(
                    self.job_id,
                    self._output_config.s3_bucket,
                    self._output_config.s3_prefix,
                    s3_client,
                ):
                    time.sleep(self._s3_polling_interval)
                    if time.time() - start > self._textract_polling_interval:
                        response = get_job_response(
                            job_id=self.job_id,
                            textract_api=TextractAPI.TextractAPI_to_Textract_API(self._api)
                            if isinstance(self._api, TextractAPI)
                            else self._api,
                            boto3_textract_client=self._textract_client,
                        )
                        job_status = response["JobStatus"]
                        if job_status == "IN_PROGRESS":
                            start = time.time()
                            response = None
                            continue
                        elif job_status == "SUCCEEDED" and "NextToken" in response:
                            response = get_full_json(
                                self.job_id,
                                TextractAPI.TextractAPI_to_Textract_API(self._api)
                                if isinstance(self._api, TextractAPI)
                                else self._api,
                                self._textract_client,
                                job_done_polling_interval=1,
                            )
                            break
                        elif job_status == "SUCCEEDED":
                            break
                        else:
                            raise Exception(f"Job failed with status: {job_status}\n{response}")
                        
                if not response:
                    response = get_full_json_from_output_config(
                        self._output_config,
                        self.job_id,
                        s3_client,
                    )
            else:
                if not self._textract_client:
                    self._textract_client = boto3.client("textract")
                response = get_full_json(
                    self.job_id,
                    TextractAPI.TextractAPI_to_Textract_API(self._api)
                    if isinstance(self._api, TextractAPI)
                    else self._api,
                    self._textract_client,
                    job_done_polling_interval=self.textract_polling_interval,
                )
            self._document = parse(response)
            if self._images is not None:
                for i, page in enumerate(self._document.pages):
                    page.image = self._images[i]
            self._document.response = response
        return object.__getattribute__(
            object.__getattribute__(self, "_document"), __name
        )



import os
from dataclasses import dataclass

from textractor.data.text_linearization_config import TextLinearizationConfig

@dataclass
class HTMLLinearizationConfig(TextLinearizationConfig):
    """
    This :class:`HTMLLinearizationConfig` is a convenience configuration for converting a Document or DocumentEntity to HTML.
    For a description of the parameters see :class:`TextLinearizationConfig`.
    """

    title_prefix: str = "<h1>"

    title_suffix: str = "</h1>"

    header_prefix: str = "<h1>"
    
    header_suffix: str = "</h1>"
    
    section_header_prefix: str = "<h2>"

    section_header_suffix: str = "</h2>"

    text_prefix: str = "<p>"

    text_suffix: str = "</p>"

    entity_layout_prefix: str = "<p>"

    entity_layout_suffix: str = "</p>"

    table_prefix: str = "<table>"

    table_suffix: str = "</table>"

    table_row_prefix: str = "<tr>"

    table_row_suffix: str = "</tr>"

    table_cell_header_prefix: str = "<th>"

    table_cell_header_suffix: str = "</th>"

    table_cell_prefix: str = "<td>"

    table_cell_suffix: str = "</td>"

    table_column_separator: str = ""

    table_linearization_format: str = "html"
    
    table_add_title_as_caption: bool = True
    
    table_add_footer_as_paragraph: bool = True

    table_column_separator: str = ""

    list_layout_prefix: str = "<div>"
    
    list_layout_suffix: str = "</div>"
    
    table_layout_prefix: str = "<div>"

    table_layout_suffix: str = "</div>"
    
    key_value_layout_prefix: str = "<div>"

    key_value_layout_suffix: str = "</div>"
    
    figure_layout_prefix: str = "<div>"
     
    figure_layout_suffix: str = "</div>"

    footer_layout_prefix: str = "<div>"
     
    footer_layout_suffix: str = "</div>"

    page_num_prefix: str = "<div>"
    
    page_num_suffix: str = "</div>"


import os
from dataclasses import dataclass


@dataclass
class TextLinearizationConfig:
    """
    The :class:`TextLinearizationConfig` object defines how a document is linearized into a text string
    """

    remove_new_lines_in_leaf_elements: bool = True  #: Removes new lines in leaf layout elements, this removes extra whitespace

    max_number_of_consecutive_new_lines: int = 2  #: Removes extra whitespace

    max_number_of_consecutive_spaces: int = None  #: Removes extra whitespace (None skips whitespace removal)

    hide_header_layout: bool = False  #: Hide headers layouts in the linearized output

    hide_footer_layout: bool = False  #: Hide footers layouts in the linearized output

    hide_figure_layout: bool = False  #: Hide figures layouts in the linearized output

    hide_table_layout: bool = False #: Hide tables layouts in the linearized output

    hide_key_value_layout: bool = False #: Hide key-value layouts in the linearized output

    hide_page_num_layout: bool = False  #: Hide page numbers in the linearized output

    page_num_prefix: str = ""  #: Prefix for page number layout elements

    page_num_suffix: str = ""  #: Suffix for page number layout elements

    same_paragraph_separator: str = (
        " "  #: Separator to use when combining elements within a text block
    )

    same_layout_element_separator: str = (
        "\n" #: Separator to use when two elements are in the same layout element
    )

    layout_element_separator: str = (
        os.linesep * 2
    )  #: Separator to use when combining linearized layout elements

    list_element_separator: str = os.linesep  #: Separator for elements in a list layout

    list_layout_prefix: str = ""  #: Prefix for list layout elements (parent)

    list_layout_suffix: str = ""  #: Suffix for list layout elements (parent)

    list_element_prefix: str = ""  #: Prefix for elements in a list layout (children)

    list_element_suffix: str = ""  #: Suffix for elements in a list layout (children)

    title_prefix: str = ""  #: Prefix for title layout elements

    title_suffix: str = ""  #: Suffix for title layout elements

    table_layout_prefix: str = os.linesep * 2  #: Prefix for table elements

    table_layout_suffix: str = os.linesep  #: Suffix for table elements

    table_remove_column_headers: bool = False  #: Remove pandas index column headers from tables

    table_column_header_threshold: float = 0.9 #: Threshold for a row to be selected as header when rendering as markdown. 0.9 means that 90% of the cells must have the is_header_cell flag. 

    table_linearization_format: str = "plaintext"  #: How to represent tables in the linearized output. Choices are plaintext, markdown or html.

    table_add_title_as_caption: bool = False #: When using html linearization format, adds the title inside the table as <caption></caption>

    # FIXME
    table_add_footer_as_paragraph: bool = False 

    table_tabulate_format: str = "github"  #: Markdown tabulate format to use when table are linearized as markdown

    table_tabulate_remove_extra_hyphens: bool = False  #: By default markdown tables will have N hyphens to preserve alignement, this reduces the number of hyphens to 1, which is the minimum number allowed by the GitHub Markdown spec

    table_duplicate_text_in_merged_cells: bool = False #: Duplicate text in merged cells to preserve line alignment

    table_flatten_headers: bool = False #: Flatten table headers into a single row, unmerging the cells horizontally

    table_min_table_words: int = 0  #: Threshold below which tables will be rendered as words instead of using table layout

    table_column_separator: str = "\t"  #: Table column separator, used when linearizing layout tables, not used if AnalyzeDocument was called with the TABLES feature

    table_flatten_semi_structured_as_plaintext: bool = False #: Ignores table structure for SEMI_STRUCTURED tables and returns them as text

    table_prefix: str = ""

    table_suffix: str = ""

    table_row_separator: str = os.linesep  #: Table row separator

    table_row_prefix: str = "" #: Prefix for table row

    table_row_suffix: str = "" #: Suffix for table row

    table_cell_prefix: str = "" #: Prefix for table cell

    table_cell_suffix: str = "" #: Suffix for table cell

    table_cell_header_prefix: str = "" #: Prefix for header cell

    table_cell_header_suffix: str = "" #: Suffix for header cell

    table_cell_empty_cell_placeholder: str = "" #: Placeholder for empty cells

    table_cell_merge_cell_placeholder: str = "" #: Placeholder for merged cell

    table_cell_left_merge_cell_placeholder: str = "" #: Placeholder for left merge cell (L) see: 

    table_cell_top_merge_cell_placeholder: str = "" #: Placeholder for left merge cell (T)

    table_cell_cross_merge_cell_placeholder: str = "" #: Placeholder for left merge cell (X)

    table_title_prefix: str = "" #: Prefix for table title if it is outside of the table (floating)
    
    table_title_suffix: str = "" #: Suffix for table title if it is outside of the table (floating)

    table_footers_prefix: str = "" #: Prefix for table footers if they are outside of the table (floating)
    
    table_footers_suffix: str = "" #: Suffix for table footers if they are outside of the table (floating)

    header_prefix: str = ""  #: Prefix for header layout elements

    header_suffix: str = ""  #: Suffix for header layout elements

    section_header_prefix: str = ""  #: Prefix for section header layout elements

    section_header_suffix: str = ""  #: Suffix for section header layout elements

    text_prefix: str = ""  #: Prefix for text layout elements

    text_suffix: str = ""  #: Suffix for text layout elements

    key_value_layout_prefix: str = ""  #: Prefix for key_value layout elements (not for individual key-value elements)

    key_value_layout_suffix: str = ""  #: Suffix for key_value layout elements (not for individual key-value elements)

    key_value_prefix: str = "" #: Prefix for key-value elements

    key_value_suffix: str = "" #: Suffix for key-value elements

    key_prefix: str = ""  #: Prefix for key elements

    key_suffix: str = " "  #: Suffix for key elements

    value_prefix: str = ""  #: Prefix for value elements

    value_suffix: str = ""  #: Suffix for value elements

    entity_layout_prefix: str = "" #: Prefix for LAYOUT_ENTITY elements (layout elements without a predicted layout type)

    entity_layout_suffix: str = "" #: Suffix for LAYOUT_ENTITY elements (layout elements without a predicted layout type)

    figure_layout_prefix: str = "" #: Prefix for figure layout elements 
    
    figure_layout_suffix: str = "" #: Suffix for figure layout elements

    footer_layout_prefix: str = "" #: Prefix for figure layout elements 
     
    footer_layout_suffix: str = "" #: Suffix for figure layout elements

    selection_element_selected: str = (
        "[X]"  #: Representation for selection element when selected
    )

    selection_element_not_selected: str = (
        "[ ]"  #: Representation for selection element when not selected
    )

    heuristic_h_tolerance: float = 0.3  #: How much the line below and above the current line should differ in width to be separated

    heuristic_line_break_threshold: float = 0.9  #: How much space is acceptable between two lines before splitting them. Expressed in multiple of min heights

    heuristic_overlap_ratio: float = 0.5  #: How much vertical overlap is tolerated between two subsequent lines before merging them into a single line

    signature_token: str = "[SIGNATURE]"  #: Signature representation in the linearized text

    add_prefixes_and_suffixes_as_words: bool = False #: Controls if the prefixes/suffixes will be inserted in the words returned by `get_text_and_words`

    add_prefixes_and_suffixes_in_text: bool = True #: Controls if the prefixes/suffixes will be added to the linearized text



"""BoundingBox class contains all the co-ordinate information for a :class:`DocumentEntity`. This class is mainly useful to locate the entity
on the image of the document page."""

from abc import ABC
import logging
import math
from typing import Tuple, List

try:
    import numpy as np
except ImportError:
    # Used in an export_as_numpy function which won't be called if the user doesn't have numpy.
    pass

from typing import Dict
from dataclasses import dataclass


class SpatialObject(ABC):
    """
    The :class:`SpatialObject` interface defines an object that has a width and height. This mostly used for :class:`BoundingBox` \
    reference to be able to provide normalized coordinates.
    """

    def __init__(self, width: float, height: float):
        self.width = width
        self.height = height


@dataclass
class BoundingBox(SpatialObject):
    """
    Represents the bounding box of an object in the format of a dataclass with (x, y, width, height). \
    By default :class:`BoundingBox` is set to work with denormalized co-ordinates: :math:`x \in [0, docwidth]` and :math:`y \in [0, docheight]`. \
    Use the as_normalized_dict function to obtain BoundingBox with normalized co-ordinates: :math:`x \in [0, 1]` and :math:`y \in [0, 1]`. \\

    Create a BoundingBox like shown below: \\

    * Directly:             :code:`bb = BoundingBox(x, y, width, height)` \\
    * From dict:            :code:`bb = BoundingBox.from_dict(bb_dict)` where :code:`bb_dict = {'x': x, 'y': y, 'width': width, 'height': height}` \\

    Use a BoundingBox like shown below: \\

    * Directly:            :code:`print('The top left is: ' + str(bb.x) + ' ' + str(bb.y))` \\
    * Convert to dict:     :code:`bb_dict = bb.as_dict()` returns :code:`{'x': x, 'y': y, 'width': width, 'height': height}`
    """

    def __init__(
        self, x: float, y: float, width: float, height: float, spatial_object=None
    ):
        super().__init__(width, height)
        self.spatial_object = spatial_object
        self.x = x
        self.y = y

    @classmethod
    def from_normalized_dict(
        cls, bbox_dict: Dict[str, float], spatial_object: SpatialObject = None
    ):
        """
        Builds an axis aligned BoundingBox from a dictionary like :code:`{'x': x, 'y': y, 'width': width, 'height': height}`. \
        The coordinates will be denormalized according to spatial_object.

        :param bbox_dict: Dictionary of normalized co-ordinates.
        :type bbox_dict: dict
        :param spatial_object: Object with width and height attributes.
        :type spatial_object: SpatialObject

        :return: Object with denormalized co-ordinates
        :rtype: BoundingBox
        """
        return cls._from_dict(bbox_dict, spatial_object)

    @staticmethod
    def _denormalize(
        x: float,
        y: float,
        width: float,
        height: float,
        spatial_object: SpatialObject = None,
    ) -> Tuple[float, float, float, float]:
        """
        Denormalizes the coordinates according to spatial_object (used as a calibrator). \
        The SpatialObject is assumed to be a container for the bounding boxes (i.e: Page). \
        Any object with width, height attributes can be used as a SpatialObject.

        :param x: Normalized co-ordinate x
        :type x: float
        :param y: Normalized co-ordinate y
        :type y: float
        :param width: Normalized width of BoundingBox
        :type width: float
        :param height: Normalized height of BoundingBox
        :type height: float
        :param spatial_object: Object with width and height attributes (i.e: Page).
        :type spatial_object: SpatialObject

        :return: Returns x, y, width, height as denormalized co-ordinates.
        :rtype: Tuple[float, float, float, float]
        """

        x = x * spatial_object.width
        y = y * spatial_object.height
        width = width * spatial_object.width
        height = height * spatial_object.height

        return x, y, width, height

    @classmethod
    def from_denormalized_xywh(
        cls,
        x: float,
        y: float,
        width: float,
        height: float,
        spatial_object: SpatialObject = None,
    ):
        """
        Builds an axis aligned bounding box from top-left, width and height properties.
        The coordinates are assumed to be denormalized.
        :param x: Left ~ [0, doc_width]
        :param y: Top  ~ [0, doc_height]
        :param width: Width ~ [0, doc_width]
        :param height: Height  ~ [0, doc_height]
        :param spatial_object: Some object with width and height attributes (i.e: Document, ConvertibleImage).
        :return: BoundingBox object in denormalized coordinates:  ~ [0, doc_height] x [0, doc_width]
        """
        return BoundingBox(x, y, width, height, spatial_object)

    @classmethod
    def from_denormalized_corners(
        cls,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        spatial_object: SpatialObject = None,
    ):
        """
        Builds an axis aligned bounding box from top-left and bottom-right coordinates.
        The coordinates are assumed to be denormalized.
        :param x1: Left  ~ [0, wdoc_idth]
        :param y1: Top ~ [0, doc_height]
        :param x2: Right  ~ [0, doc_width]
        :param y2: Bottom ~ [0, doc_height]
        :param spatial_object: Some object with width and height attributes (i.e: Document, ConvertibleImage).
        :return: BoundingBox object in denormalized coordinates:  ~ [0, doc_height] x [0, doc_width]
        """
        x = x1
        y = y1
        width = x2 - x1
        height = y2 - y1
        return BoundingBox(x, y, width, height, spatial_object=spatial_object)

    @classmethod
    def from_denormalized_borders(
        cls,
        left: float,
        top: float,
        right: float,
        bottom: float,
        spatial_object: SpatialObject = None,
    ):
        """
        Builds an axis aligned bounding box from top-left and bottom-right coordinates.
        The coordinates are assumed to be denormalized.
        If spatial_object is not None, the coordinates will be denormalized according to the spatial object.
        :param left: ~ [0, doc_width]
        :param top: ~ [0, doc_height]
        :param right: ~ [0, doc_width]
        :param bottom: ~ [0, doc_height]
        :param spatial_object: Some object with width and height attributes
        :return: BoundingBox object in denormalized coordinates:  ~ [0, doc_height] x [0, doc_width]
        """
        return cls.from_denormalized_corners(left, top, right, bottom, spatial_object)

    @classmethod
    def from_denormalized_dict(cls, bbox_dict: Dict[str, float]):
        """
        Builds an axis aligned bounding box from a dictionary of:
        {'x': x, 'y': y, 'width': width, 'height': height}
        The coordinates will be denormalized according to the spatial object.
        :param bbox_dict: {'x': x, 'y': y, 'width': width, 'height': height} of [0, doc_height] x [0, doc_width]
        :param spatial_object: Some object with width and height attributes
        :return: BoundingBox object in denormalized coordinates:  ~ [0, doc_height] x [0, doc_width]
        """
        x = bbox_dict["x"]
        y = bbox_dict["y"]
        width = bbox_dict["width"]
        height = bbox_dict["height"]
        return BoundingBox(x, y, width, height)

    @classmethod
    def enclosing_bbox(cls, bboxes, spatial_object: SpatialObject = None):
        """
        :param bboxes [BoundingBox]: list of bounding boxes
        :param spatial_object SpatialObject: spatial object to be added to the returned bbox
        :return:
        """
        bboxes = [bbox for bbox in bboxes if bbox is not None]
        if bboxes and not isinstance(bboxes[0], BoundingBox):
            try:
                bboxes = [bbox.bbox for bbox in bboxes]
            except:
                raise Exception(
                    "bboxes must be of type List[BoundingBox] or of type List[DocumentEntity]"
                )

        x1, y1, x2, y2 = float("inf"), float("inf"), float("-inf"), float("-inf")\
        
        # FIXME: This should not happen
        if not any([bbox is not None for bbox in bboxes]):
            logging.warning("At least one bounding box needs to be non-null")
            return BoundingBox(0, 0, 1, 1, spatial_object=spatial_object)            
        
        if spatial_object is None:
            spatial_object = bboxes[0].spatial_object

        for bbox in bboxes:
            if bbox is not None:
                x1 = min(x1, bbox.x)
                x2 = max(x2, bbox.x + bbox.width)
                y1 = min(y1, bbox.y)
                y2 = max(y2, bbox.y + bbox.height)
        return BoundingBox(x1, y1, x2 - x1, y2 - y1, spatial_object=spatial_object)

    @classmethod
    def is_inside(cls, bbox_a, bbox_b):
        """Returns true if Bounding Box A is within Bounding Box B
        """
        return (
            bbox_a.x >= bbox_b.x and
            (bbox_a.x + bbox_a.width) <= (bbox_b.x + bbox_b.width) and
            bbox_a.y >= bbox_b.y and
            (bbox_a.y + bbox_a.height) <= (bbox_b.y + bbox_b.height)
        )

    @classmethod
    def center_is_inside(cls, bbox_a, bbox_b):
        """Returns true if the center point of Bounding Box A is within Bounding Box B
        """
        return (
            (bbox_a.x + bbox_a.width / 2) >= bbox_b.x and
            (bbox_a.x + bbox_a.width / 2) <= (bbox_b.x + bbox_b.width) and
            (bbox_a.y + bbox_a.height / 2) >= bbox_b.y and
            (bbox_a.y + bbox_a.height / 2) <= (bbox_b.y + bbox_b.height)
        )

    @property
    def area(self):
        """
        Returns the area of the bounding box, handles negative bboxes as 0-area

        :return: Bounding box area
        :rtype: float
        """
        if self.width < 0 or self.height < 0:
            return 0
        else:
            return self.width * self.height

    @classmethod
    def _from_dict(
        cls, bbox_dict: Dict[str, float], spatial_object: SpatialObject = None
    ):
        """
        Builds an axis aligned BoundingBox from a dictionary like :code:`{'x': x, 'y': y, 'width': width, 'height': height}`. \
        The co-ordinates will be denormalized according to spatial_object.

        :param bbox_dict: Dictionary of normalized co-ordinates.
        :type bbox_dict: dict
        :param spatial_object: Object with width and height attributes.
        :type spatial_object: SpatialObject

        :return: Object with denormalized coordinates
        :rtype: BoundingBox
        """
        x = bbox_dict["Left"]
        y = bbox_dict["Top"]
        width = bbox_dict["Width"]
        height = bbox_dict["Height"]
        # if spatial_object is not None:
        #    x, y, width, height = cls._denormalize(x, y, width, height, spatial_object)
        # else:
        #    spatial_object = None
        return BoundingBox(x, y, width, height, spatial_object)

    def as_denormalized_numpy(self):
        """
        :return: Returns denormalized co-ordinates x, y and dimensions width, height as numpy array.
        :rtype: numpy.array
        """
        return np.array([self.x, self.y, self.width, self.height])

    def get_intersection(self, bbox):
        """
        Returns the intersection of this object's bbox and another BoundingBox
        :return: a BoundingBox object
        """
        assert isinstance(bbox, BoundingBox)
        x1y1x2y2 = (
            max(self.x, bbox.x),
            max(self.y, bbox.y),
            min(self.x + self.width, bbox.x + bbox.width),
            min(self.y + self.height, bbox.y + bbox.height),
        )
        return BoundingBox.from_denormalized_corners(
            *x1y1x2y2, spatial_object=self.spatial_object
        )

    def get_distance(self, bbox):
        """
        Returns the distance between the center point of the bounding box and another bounding box

        :return: Returns the distance as float
        :rtype: float
        """
        return math.sqrt(
            ((self.x + self.width / 2) - (bbox.x + bbox.width / 2)) ** 2
            + ((self.y + self.height / 2) - (bbox.y + bbox.height / 2)) ** 2
        )

    def __repr__(self):
        return f"x: {self.x}, y: {self.y}, width: {self.width}, height: {self.height}"



"""
Represents a single :class:`SelectionElement`/Checkbox/Clickable Entity within the :class:`Document`.

This class contains the associated metadata with the :class:`SelectionElement` entity including the entity ID, 
bounding box information, selection status, page number, Page ID and confidence of detection.
"""

import uuid

from typing import List
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.word import Word
from textractor.entities.line import Line
from textractor.entities.bbox import BoundingBox
from textractor.data.constants import SELECTED, NOT_SELECTED, SelectionStatus
from textractor.entities.document_entity import DocumentEntity


class SelectionElement(DocumentEntity):
    """
    To create a new :class:`SelectionElement` object we need the following:

    :param entity_id: Unique identifier of the SelectionElement entity.
    :type entity_id: str
    :param bbox: Bounding box of the SelectionElement
    :type bbox: BoundingBox
    :param status: SelectionStatus.SELECTED / SelectionStatus.NOT_SELECTED
    :type status: SelectionStatus
    :param confidence: Confidence with which this entity is detected.
    :type confidence: float
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        status: SelectionStatus,
        confidence: float = 0,
    ):
        super().__init__(entity_id, bbox)
        self.key_id = None
        self.value_id = None
        self.status = status
        self._confidence = confidence / 100
        self._page = None
        self._page_id = None

    def is_selected(self) -> bool:
        """
        :return: Returns True / False depending on selection status of the SelectionElement.
        :rtype: bool
        """
        return self.status == SelectionStatus.SELECTED

    @property
    def words(self) -> List[Word]:
        """
        :return: Empty Word list as SelectionElement do not have words
        :rtype: EntityList[Word]
        """
        return []

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        w = Word(
            entity_id=str(uuid.uuid4()),
            bbox=self.bbox,
            text=config.selection_element_selected
            if self.status == SelectionStatus.SELECTED
            else config.selection_element_not_selected,
        )
        w.is_clickable = True
        w.line = Line(entity_id=str(uuid.uuid4()), bbox=self.bbox, words=[w])

        words = [w]

        text = w.text
            
        for w in words:
            w.value_id = str(self.id)
            w.value_bbox = self.bbox
        return text, words

    @property
    def page(self):
        """
        :return: Returns the page number of the page the SelectionElement entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the SelectionElement entity.

        :param page_num: Page number where the SelectionElement entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the SelectionElement entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    def __repr__(self) -> str:
        """
        Returns string representation of SelectionElement.
        """
        if self.status == SelectionStatus.SELECTED:
            return "[X]"
        else:
            return "[ ]"



"""
:class:`Textractor` is the main class associated with this package. It needs to be instantiated before using any of the functionalities
the package provides. The main use of this class is to make calls to the Textract API and create Python objects for all the
document entities that are returned in the JSON output of the API. The response received is implicitly parsed and a :class:`Document` type 
object is returned containing all the document entities, their associated relationships and metadata.

The Textract API and Textractor method mapping is as below. Use these wrappers to make calls and parse the responses
in one step.

* (SYNC) DetectDocumentText : detect_document_text
* (SYNC) AnalyzeDocument : analyze_document
* (SYNC) AnalyzeID : analyze_id
* (SYNC) AnalyzeExpense : analyze_expense
* (ASYNC) StartDocumentTextDetection : start_document_text_detection
* (ASYNC) StartDocumentAnalysis : start_document_analysis
* (ASYNC) StartExpenseAnalysis : start_expense_analysis

"""

import io
import os
import boto3
import logging
import uuid
from PIL import Image
from copy import deepcopy
from typing import List, Union
from textractcaller import (
    call_textract,
    call_textract_analyzeid,
    call_textract_expense,
    OutputConfig,
    Query,
    QueriesConfig,
)
from textractcaller.t_call import Textract_Call_Mode, Textract_API, get_full_json

try:
    try:
        import pypdfium2
    except:
        import pdf2image

    IS_PDF_RENDERING_ENABLED = True
except ImportError:
    IS_PDF_RENDERING_ENABLED = False
    logging.info("pypdfium2 and pdf2image are both not installed, client-side PDF rasterizing is disabled")

from textractor.data.constants import (
    TextractAPI,
    TextractFeatures,
)
from textractor.entities.document import Document
from textractor.entities.lazy_document import LazyDocument
from textractor.parsers import response_parser
from textractor.utils.s3_utils import upload_to_s3, s3_path_to_bucket_and_prefix
from textractor.utils.pdf_utils import rasterize_pdf
from textractor.exceptions import (
    InputError,
    IncorrectMethodException,
    MissingDependencyException,
    UnhandledCaseException,
    UnsupportedDocumentException,
    InvalidS3ObjectException,
)


class Textractor:
    """
    Initializes the customer credentials needed to make calls to Textract using boto3 package internally.

    :param profile_name: Customer's profile name as set in the ~/.aws/config file. This profile typically contains this format.
                                :code:`[default]
                                region = us-west-2
                                output=json`
    :type profile_name: str
    :param region_name: If AWSCLI isn't setup, the user can pass region to let boto3 pick up credentials from the system.
    :param region_name: str
    :type profile_name: str, optional
    :param kms_key_id: Customer's AWS KMS key (cryptographic key)
    :type kms_key_id: str, optional
    """

    def __init__(
        self,
        profile_name: str = None,
        region_name: str = None,
        kms_key_id: str = "",
    ):
        self.profile_name = profile_name
        self.region_name = region_name
        self.kms_key_id = kms_key_id

        if self.profile_name is not None:
            self.session = boto3.session.Session(profile_name=self.profile_name)
        elif self.region_name is not None:
            self.session = boto3.session.Session(region_name=self.region_name)
        elif os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION"):
            # We support both AWS_REGION and AWS_DEFAULT_REGION, with AWS_REGION having precedence.
            self.region_name = os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION")
            self.session = boto3.session.Session(region_name=self.region_name)
        else:
            raise InputError(
                "Unable to initiate Textractor. Either profile_name or region requires an input parameter."
            )
        if self.region_name is not None:
            self.textract_client = self.session.client(
                "textract", region_name=self.region_name
            )
        else:
            self.textract_client = self.session.client("textract")
        self.s3_client = self.session.client("s3")

    def _get_document_images_from_path(self, filepath: str) -> List[Image.Image]:
        """
        Converts the every page in the document to an image. It supports pdfs and image formats that can be opened by
        PIL package. Documents can be stored in the local computer or on an S3 Bucket.

        :param filepath: filepath to the document stored locally or on an S3 bucket.
        :type filepath: str, required
        :return: Returns a list of PIL Images, one for each page of the document
        :rtype: List[Image]
        """
        images = []
        if "s3://" in filepath:
            edit_filepath = filepath.replace("s3://", "")
            bucket = edit_filepath.split("/")[0]
            key = edit_filepath[edit_filepath.index("/") + 1 :]

            s3_client = (
                boto3.session.Session(profile_name=self.profile_name).client("s3")
                if self.profile_name is not None
                else boto3.session.Session(region_name=self.region_name).client("s3")
            )
            file_obj = s3_client.get_object(Bucket=bucket, Key=key).get("Body").read()
            if filepath.lower().endswith(".pdf"):
                if IS_PDF_RENDERING_ENABLED:
                    images = rasterize_pdf(bytearray(file_obj))
                else:
                    raise MissingDependencyException(
                        "pdf2image is not installed. If you do not plan on using visualizations you can skip image generation using save_image=False in your function call."
                    )
            else:
                images = [Image.open(io.BytesIO(bytearray(file_obj)))]

        else:
            if filepath.lower().endswith(".pdf"):
                if IS_PDF_RENDERING_ENABLED:
                    images = rasterize_pdf(filepath)
                else:
                    raise MissingDependencyException(
                        "pdf2image is not installed. If you do not plan on using visualizations you can skip image generation using save_image=False in your function call."
                    )
            else:
                images = [Image.open(filepath)]

        if not images:
            raise UnhandledCaseException(f"Could not get any images from {filepath}")

        return images

    def detect_document_text(
        self, file_source, save_image: bool = True
    ) -> Document:
        """
        Make a call to the SYNC DetectDocumentText API, implicitly parses the response and produces a :class:`Document` object.
        This function is ideal for single page PDFs or single images.

        :param file_source: Path to a file stored locally, on an S3 bucket or PIL Image
        :type file_source: str or PIL.Image, required
        :param save_image: Flag to indicate if document images are to be stored within the Document object. This is optional
                            and necessary only if the customer wants to visualize bounding boxes for their document entities.
        :type save_image: bool

        :return: Returns a Document object containing all the entities, relationships and metadata extracted by the Textract
                 DetectDocumentText API stored within it.
        :rtype: Document
        """

        if isinstance(file_source, list) and len(file_source) > 1:
            raise IncorrectMethodException(
                "List contains more than 1 image. Call start_document_text_detection instead."
            )

        elif isinstance(file_source, str):
            logging.debug("Filepath given.")
            if not save_image and file_source.lower().endswith(".pdf"):
                images = []
            else:
                images = self._get_document_images_from_path(file_source)
                if len(images) > 1:
                    raise IncorrectMethodException(
                        "Input contains more than 1 page. Call start_document_analysis() instead."
                    )
                file_source = _image_to_byte_array(images[0])    

        elif isinstance(file_source, Image.Image):
            logging.debug("PIL Image given.")
            images = [file_source]
            file_source = _image_to_byte_array(file_source)

        elif isinstance(file_source, list) and isinstance(file_source[0], Image.Image):
            logging.debug("List of PIL Image given.")
            images = deepcopy(file_source)
            file_source = _image_to_byte_array(images[0])

        else:
            images = []
            raise InputError("Input file_source format not supported.")

        try:
            response = call_textract(
                input_document=file_source,
                features=[],
                queries_config=None,  # not supported yet
                output_config=None,
                kms_key_id=self.kms_key_id,
                job_tag="",
                notification_channel=None,  # not supported yet
                client_request_token="",
                return_job_id=False,
                force_async_api=False,
                call_mode=Textract_Call_Mode.FORCE_SYNC,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=0,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            elif exception.__class__.__name__ == "UnsupportedDocumentException":
                raise UnsupportedDocumentException(
                    "Textract returned UnsupportedDocumentException, if file_source is a PDF, make sure that it only has one page or use start_document_text_detection. If your file_source is an image, make sure that it is not larger than 5MB."
                )
            raise exception

        document = response_parser.parse(response)
        document.response = response
        if save_image:
            for page in document.pages:
                page.image = images[document.pages.index(page)]
        return document

    def start_document_text_detection(
        self,
        file_source: Union[str, bytes, Image.Image],
        s3_output_path: str = "",
        s3_upload_path: str = "",
        client_request_token: str = "",
        job_tag: str = "",
        save_image: bool = True,
    ) -> LazyDocument:
        """
        Make a call to the ASYNC StartDocumentTextDetection API.

        :param file_source: File bytes, path to a file stored locally or in an S3 bucket
        :type file_source: Union[str, bytes, Image.Image], required
        :param s3_output_path: Prefix to store the output on the S3 bucket (passed as param to Textractor).
        :type s3_output_path: str
        :param s3_upload_path: If given, will automatically upload the document to the given S3 prefix before calling Textract. Files are uploaded
                                    under a uuid. If not given the data is expected to be already in s3
        :type s3_upload_path: str, optional
        :param client_request_token: The idempotent token that's used to identify the start request. If you use the same. token
                                    with multiple StartDocumentTextDetection requests, the same. JobId is returned. Use ClientRequestToken
                                    to prevent the same. job from being accidentally started more than once.
        :type client_request_token: str, optional
        :param job_tag: An identifier that you specify that's included in the completion notification published to the Amazon SNS topic.
        :type job_tag: str, optional
        :param save_image: Flag to indicate if document images are to be stored within the Document object. This is optional
                            and necessary only if the customer wants to visualize bounding boxes for their document entities.
        :type save_image: bool

        :return: Lazy-loaded Document object
        :rtype: LazyDocument
        """

        original_file_source = file_source

        if not isinstance(file_source, (str, bytes, Image.Image)):
            raise InputError(
                f"file_source needs to be of type str, bytes or PIL Image, not {type(file_source)}"
            )

        # If the file is not already in S3
        if not isinstance(file_source, str) or not file_source.startswith("s3://"):
            # Check if the user has given us a bucket to upload to
            if not s3_upload_path:
                raise InputError(
                    "For files not in S3, an S3 upload path must be provided"
                )

            s3_file_path = os.path.join(s3_upload_path, str(uuid.uuid4()))
            upload_to_s3(self.s3_client, s3_file_path, file_source)
            file_source = s3_file_path

        output_config = None
        if s3_output_path:
            s3_bucket, s3_prefix = s3_path_to_bucket_and_prefix(s3_output_path)
            output_config = OutputConfig(s3_bucket=s3_bucket, s3_prefix=s3_prefix)

        try:
            response = call_textract(
                input_document=file_source,
                features=[],
                queries_config=None,  # not supported yet
                output_config=output_config,
                kms_key_id=self.kms_key_id,
                job_tag=job_tag,
                notification_channel=None,  # not supported yet
                client_request_token=client_request_token,
                return_job_id=True,
                force_async_api=True,
                call_mode=Textract_Call_Mode.FORCE_ASYNC,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=1,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            raise exception

        images = None
        if save_image:
            if isinstance(original_file_source, Image.Image):
                images = [original_file_source]
            elif (
                isinstance(original_file_source, list)
                and len(original_file_source)
                and isinstance(original_file_source[0], Image.Image)
            ):
                images = original_file_source
            else:
                images = self._get_document_images_from_path(original_file_source)

        return LazyDocument(
            response["JobId"],
            TextractAPI.DETECT_TEXT,
            textract_client=self.textract_client,
            images=images,
        )

    def analyze_document(
        self,
        file_source,
        features,
        queries: Union[QueriesConfig, List[Query], List[str]] = None,
        save_image: bool = True,
    ) -> Document:
        """
        Make a call to the SYNC AnalyzeDocument API, implicitly parses the response and produces a :class:`Document` object.
        This function is ideal for single page PDFs or single images.

        :param file_source: Path to a file stored locally, on an S3 bucket or PIL Image
        :type file_source: str or PIL.Image, required
        :param features: List of TextractFeatures to be extracted from the Document by the TextractAPI
        :type features: list, required
        :param queries: Queries to run on the document
        :type features: Union[QueriesConfig, List[Query], List[str]]
        :param save_image: Flag to indicate if document images are to be stored within the Document object. This is optional
                            and necessary only if the customer wants to visualize bounding boxes for their document entities.
        :type save_image: bool

        :return: Returns a Document object containing all the entities, relationships and metadata extracted by the Textract
                 AnalyzeDocument API stored within it.
        :rtype: Document
        """
        if isinstance(file_source, list) and len(file_source) > 1:
            raise IncorrectMethodException(
                "List contains more than 1 image. Call start_document_analysis() instead."
            )

        elif isinstance(file_source, str):
            logging.debug("Filepath given.")
            if not save_image and file_source.lower().endswith(".pdf"):
                images = []
            else:
                images = self._get_document_images_from_path(file_source)
                if len(images) > 1:
                    raise IncorrectMethodException(
                        "Input contains more than 1 page. Call start_document_analysis() instead."
                    )
                file_source = _image_to_byte_array(images[0])    
        elif isinstance(file_source, Image.Image):
            logging.debug("PIL Image given.")
            images = [file_source]
            file_source = _image_to_byte_array(file_source)

        elif isinstance(file_source, list) and isinstance(file_source[0], Image.Image):
            logging.debug("List of PIL Image given.")
            images = deepcopy(file_source)
            file_source = _image_to_byte_array(images[0])

        else:
            images = []
            raise InputError("Input file_source format not supported.")

        if not isinstance(features, list):
            features = [features]

        if queries and TextractFeatures.QUERIES not in features:
            raise InputError(
                "Queries were given as a parameter but QUERIES is not enabled in the feature set"
            )

        if queries and not isinstance(queries, QueriesConfig):
            if not isinstance(queries, List):
                raise InputError(
                    f"Queries must be of type QueriesConfig, List[Query] or List[str], not {type(queries)}"
                )
            if isinstance(queries[0], Query):
                queries_config = QueriesConfig(queries)
                queries = queries_config
            elif isinstance(queries[0], str):
                queries_config = QueriesConfig([Query(query) for query in queries])
                queries = queries_config
            else:
                raise InputError(
                    f"Queries must be of type QueriesConfig, List[Query] or List[str], not {type(queries)}"
                )

        try:
            response = call_textract(
                input_document=file_source,
                features=features,
                queries_config=queries,  # not supported yet
                output_config=None,
                kms_key_id=self.kms_key_id,
                job_tag="",
                notification_channel=None,  # not supported yet
                client_request_token="",
                return_job_id=False,
                force_async_api=False,
                call_mode=Textract_Call_Mode.FORCE_SYNC,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=0,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            elif exception.__class__.__name__ == "UnsupportedDocumentException":
                raise UnsupportedDocumentException(
                    "Textract returned an UnsupportedDocumentException, if file_source is a PDF, make sure that it only has one page or use start_document_analysis. If your file_source is an image, make sure that it is not larger than 5MB."
                )
            raise exception

        document = response_parser.parse(response)
        document.response = response
        if save_image:
            for page in document.pages:
                page.image = images[document.pages.index(page)]
        return document

    def start_document_analysis(
        self,
        file_source: Union[str, bytes, Image.Image],
        features,
        s3_output_path: str = "",
        s3_upload_path: str = "",
        queries: Union[QueriesConfig, List[Query], List[str]] = None,
        client_request_token: str = "",
        job_tag: str = "",
        save_image: bool = True,
    ) -> LazyDocument:
        """
        Make a call to the ASYNC StartDocumentAnalysis API, implicitly parses the response and produces a :class:`Document` object.
        This function is ideal for multipage PDFs or an image.

        :param file_source: Path to a file stored locally, on an S3 bucket or a PIL Image
        :type file_source: Union[str, bytes, Image.Image], required
        :param features: List of TextractFeatures to be extracted from the Document by the TextractAPI
        :type features: list, required
        :param s3_output_path: Path to store the output on the S3 bucket (passed as param to Textractor).
        :type s3_output_path: str
        :param s3_upload_path: If given, will automatically upload the document to the given S3 prefix before calling Textract. Files are uploaded
                               under a uuid. If not given the data is expected to be already in s3
        :type s3_upload_path: str, optional
        :param client_request_token: The idempotent token that's used to identify the start request. If you use the same. token
                                    with multiple StartDocumentTextDetection requests, the same. JobId is returned. Use ClientRequestToken
                                    to prevent the same. job from being accidentally started more than once.
        :type client_request_token: str, optional
        :param job_tag: An identifier that you specify that's included in the completion notification published to the Amazon SNS topic.
        :type job_tag: str, optional
        :param save_image: Flag to indicate if document images are to be stored within the Document object. This is optional
                            and necessary only if the customer wants to visualize bounding boxes for their document entities.
        :type save_image: bool

        :return: Returns a Document object containing all the entities, relationships and metadata extracted by the Textract
                 StartDocumentAnalysis API stored within it.
        :rtype: Document
        """

        original_file_source = file_source

        if not isinstance(file_source, (str, bytes, Image.Image)):
            raise InputError(
                f"file_source needs to be of type str, bytes or PIL Image, not {type(file_source)}"
            )

        # If the file is not already in S3
        if not isinstance(file_source, str) or not file_source.startswith("s3://"):
            # Check if the user has given us a bucket to upload to
            if not s3_upload_path:
                raise InputError(
                    f"For files not in S3, an S3 upload path must be provided"
                )

            s3_file_path = os.path.join(s3_upload_path, str(uuid.uuid4()))
            upload_to_s3(self.s3_client, s3_file_path, file_source)
            file_source = s3_file_path

        output_config = None
        if s3_output_path:
            s3_bucket, s3_prefix = s3_path_to_bucket_and_prefix(s3_output_path)
            output_config = OutputConfig(s3_bucket=s3_bucket, s3_prefix=s3_prefix)

        if not isinstance(features, list):
            features = [features]

        if queries and TextractFeatures.QUERIES not in features:
            raise InputError(
                "Queries were given as a parameter but QUERIES is not enabled in the feature set"
            )

        if queries and not isinstance(queries, QueriesConfig):
            if not isinstance(queries, List):
                raise InputError(
                    f"Queries must be of type QueriesConfig, List[Query] or List[str], not {type(queries)}"
                )
            if isinstance(queries[0], Query):
                queries_config = QueriesConfig(queries)
                queries = queries_config
            elif isinstance(queries[0], str):
                queries_config = QueriesConfig([Query(query) for query in queries])
                queries = queries_config
            else:
                raise InputError(
                    f"Queries must be of type QueriesConfig, List[Query] or List[str], not {type(queries)}"
                )

        try:
            response = call_textract(
                input_document=file_source,
                features=features,
                queries_config=queries,  # not supported yet
                output_config=output_config,
                kms_key_id=self.kms_key_id,
                job_tag=job_tag,
                notification_channel=None,  # not supported yet
                client_request_token=client_request_token,
                return_job_id=True,
                force_async_api=True,
                call_mode=Textract_Call_Mode.FORCE_ASYNC,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=1,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            raise exception

        images = None
        if save_image:
            if isinstance(original_file_source, Image.Image):
                images = [original_file_source]
            elif (
                isinstance(original_file_source, list)
                and len(original_file_source)
                and isinstance(original_file_source[0], Image.Image)
            ):
                images = original_file_source
            else:
                images = self._get_document_images_from_path(original_file_source)

        return LazyDocument(
            response["JobId"],
            TextractAPI.ANALYZE,
            textract_client=self.textract_client,
            images=images,
            output_config=output_config,
        )

    def analyze_id(
        self,
        file_source: Union[str, List[Image.Image], List[str]],
        save_image: bool = True,
    ) -> Document:
        """AnalyzeID parses identity documents such as passports and driver's license and
        returns the result as a dictionary of standardized fields. See https://docs.aws.amazon.com/textract/latest/dg/identitydocumentfields.html
        for a complete list.

        :param file_source: Path to a file stored locally, on an S3 bucket or list of PIL Images
        :type file_source: Union[str, List[Image.Image], List[str]]
        :param save_image: Saves the images in the returned Document object for visualizing the results, defaults to False
        :type save_image: bool, optional
        :raises InputError: Raised when the file_source could not be parsed
        :raises InvalidS3ObjectException: Raised when the S3 object passed as file source is in a region that does not match the one used to create the Textractor object.
        :raises exception: Raised when the Textract call fails
        :return: Document
        :rtype: Document
        """
        if isinstance(file_source, str):
            logging.debug("Filepath given.")
            images = self._get_document_images_from_path(file_source)
        elif isinstance(file_source, Image.Image):
            logging.debug("PIL Image given.")
            images = [file_source]
        elif isinstance(file_source, list) and isinstance(file_source[0], Image.Image):
            logging.debug("List of PIL Image given.")
            # FIXME: Is this needed?
            images = deepcopy(file_source)
        else:
            images = []
            raise InputError("Input file_source format not supported.")

        images_bytes = [_image_to_byte_array(image) for image in images]

        try:
            response = call_textract_analyzeid(
                document_pages=images_bytes,
                boto3_textract_client=self.textract_client,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            raise exception

        document = response_parser.parse(response)
        document.response = response
        if save_image:
            for page in document.pages:
                page.image = images[document.pages.index(page)]
        return document

    def analyze_expense(
        self,
        file_source: Union[str, List[Image.Image], List[str]],
        save_image: bool = True,
    ):
        """Make a call to the SYNC AnalyzeExpense API, implicitly parses the response and produces a :class:`Document` object.
        This function is ideal for multipage PDFs or list of images.

        :param file_source: Path to a file stored locally, on an S3 bucket or PIL Image
        :type file_source: Union[str, List[Image.Image], List[str]]
        :param save_image: Whether to keep the file source as PIL Images inside the returned Document object, defaults to False
        :type save_image: bool, optional
        :raises IncorrectMethodException: Raised when the file source type is incompatible with the Textract API being called
        :raises InputError: Raised when the file source type is invalid
        :raises InvalidS3ObjectException: Raised when the file source region is different the API region.
        :raises exception: Raised if the Textract API call fails
        :return: Document
        :rtype: Document
        """
        if isinstance(file_source, list) and len(file_source) > 1:
            raise IncorrectMethodException(
                "List contains more than 1 image. Call start_expense_analysis instead."
            )

        elif isinstance(file_source, str):
            logging.debug("Filepath given.")
            images = self._get_document_images_from_path(file_source)
            if len(images) > 1:
                raise IncorrectMethodException(
                    "Input contains more than 1 page. Call start_expense_analysis instead."
                )
            file_source = _image_to_byte_array(images[0])

        elif isinstance(file_source, Image.Image):
            logging.debug("PIL Image given.")
            images = [file_source.copy()]
            file_source = _image_to_byte_array(file_source)

        elif isinstance(file_source, list) and isinstance(file_source[0], Image.Image):
            logging.debug("List of PIL Image given.")
            images = deepcopy(file_source)
            file_source = _image_to_byte_array(images[0])

        else:
            images = []
            raise InputError("Input file_source format not supported.")

        output_config = None

        try:
            response = call_textract_expense(
                input_document=file_source,
                output_config=output_config,
                kms_key_id=self.kms_key_id,
                job_tag="",
                notification_channel=None,  # not supported yet
                client_request_token="",
                return_job_id=False,
                force_async_api=False,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=0,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            raise exception

        document = response_parser.parse(response)
        document.response = response
        if save_image:
            for page in document.pages:
                page.image = images[document.pages.index(page)]
        return document

    def start_expense_analysis(
        self,
        file_source: Union[str, bytes, Image.Image],
        s3_output_path: str = "",
        s3_upload_path: str = "",
        client_request_token: str = "",
        job_tag: str = "",
        save_image: bool = True,
    ) -> LazyDocument:
        """Make a call to the ASYNC StartExpenseAnalysis API, implicitly parses the response and produces a :class:`Document` object.
        This function is ideal for multipage PDFs or an image.

        :param file_source: Path to a file stored locally, on an S3 bucket or a PIL Image
        :type file_source: Union[str, bytes, Image.Image]
        :param s3_output_path: Path to store the output on the S3 bucket (passed as param to Textractor).
        :type s3_output_path: str
        :param s3_upload_path: If given, will automatically upload the document to the given S3 prefix before calling Textract. Files are uploaded
                               under a uuid. If not given the data is expected to be already in s3
        :type s3_upload_path: str, optional
        :param client_request_token: The idempotent token that's used to identify the start request. If you use the same. token
                                    with multiple StartDocumentTextDetection requests, the same. JobId is returned. Use ClientRequestToken
                                    to prevent the same. job from being accidentally started more than once.
        :type client_request_token: str, optional
        :param job_tag: An identifier that you specify that's included in the completion notification published to the Amazon SNS topic.
        :type job_tag: str, optional
        :param save_image: Flag to indicate if document images are to be stored within the Document object. This is optional
                            and necessary only if the customer wants to visualize bounding boxes for their document entities.
        :type save_image: bool
        :raises InputError: Raised when the file source type is invalid
        :raises InvalidS3ObjectException: Raised when the file source region is different the API region.
        :raises exception: Raised if the Textract API call fails
        :return: Lazy-loaded Document object
        :rtype: LazyDocument
        """

        original_file_source = file_source

        if not isinstance(file_source, (str, bytes, Image.Image)):
            raise InputError(
                f"file_source needs to be of type str, bytes or PIL Image, not {type(file_source)}"
            )

        # If the file is not already in S3
        if not isinstance(file_source, str) or not file_source.startswith("s3://"):
            # Check if the user has given us a bucket to upload to
            if not s3_upload_path:
                raise InputError(
                    f"For files not in S3, an S3 upload path must be provided"
                )

            s3_file_path = os.path.join(s3_upload_path, str(uuid.uuid4()))
            upload_to_s3(self.s3_client, s3_file_path, file_source)
            file_source = s3_file_path

        output_config = None
        if s3_output_path:
            s3_bucket, s3_prefix = s3_path_to_bucket_and_prefix(s3_output_path)
            output_config = OutputConfig(s3_bucket=s3_bucket, s3_prefix=s3_prefix)

        try:
            response = call_textract_expense(
                input_document=file_source,
                output_config=output_config,
                kms_key_id=self.kms_key_id,
                job_tag=job_tag,
                notification_channel=None,  # not supported yet
                client_request_token=client_request_token,
                return_job_id=True,
                force_async_api=True,
                boto3_textract_client=self.textract_client,
                job_done_polling_interval=1,
            )
        except Exception as exception:
            if exception.__class__.__name__ == "InvalidS3ObjectException":
                raise InvalidS3ObjectException(
                    "Textract returned InvalidS3ObjectException. Ensure that the s3 path is correct and that both the Textract API and the bucket are in the same region."
                )
            raise exception

        images = None
        if save_image:
            if isinstance(original_file_source, Image.Image):
                images = [original_file_source]
            elif (
                isinstance(original_file_source, list)
                and len(original_file_source)
                and isinstance(original_file_source[0], Image.Image)
            ):
                images = original_file_source
            else:
                images = self._get_document_images_from_path(original_file_source)

        return LazyDocument(
            response["JobId"],
            TextractAPI.EXPENSE,
            textract_client=self.textract_client,
            images=images,
        )

    def get_result(
        self, job_id: str, api: Union[TextractAPI, Textract_API]
    ) -> Document:
        """
        Retrieves Textract API output for a given job id.
        :param job_id: Textract API JobID
        :type job_id: str, required
        :return: Returns a Document object
        :rtype: Document
        """

        response = get_full_json(
            job_id,
            TextractAPI.TextractAPI_to_Textract_API(api)
            if isinstance(api, TextractAPI)
            else api,
            boto3_textract_client=self.textract_client,
            job_done_polling_interval=1,
        )

        document = response_parser.parse(response)
        document.response = response

        return document


def _image_to_byte_array(image: Image) -> bytes:
    """
    Function to convert PIL.Image to bytearray for processing Document using Textract service.
    :param image: Image to be converted to bytearray
    :type image: PIL.Image, required
    :return: Returns a bytearray of the input image
    :rtype: bytes
    """
    img_byte_arr = io.BytesIO()
    # We set quality to 95 and subsampling to 0 because the pillow defaults are very low resolution
    image.convert("RGB").save(img_byte_arr, format="JPEG", quality=95, subsampling=0)
    img_byte_arr = img_byte_arr.getvalue()
    return img_byte_arr



"""
Represents a single :class:`Word` within the :class:`Document`.
This class contains the associated metadata with the :class:`Word` entity including the text transcription, 
text type, bounding box information, page number, Page ID and confidence of detection.
"""

from textractor.data.constants import TextTypes
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity


class Word(DocumentEntity):
    """
    To create a new :class:`Word` object we need the following:

    :param entity_id: Unique identifier of the Word entity.
    :type entity_id: str
    :param bbox: Bounding box of the Word entity.
    :type bbox: BoundingBox
    :param text: Transcription of the Word object.
    :type text: str
    :param text_type: Enum value stating the type of text stored in the entity. Takes 2 values - PRINTED and HANDWRITING
    :type text_type: TextTypes
    :param confidence: value storing the confidence of detection out of 100.
    :type confidence: float
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        text: str = "",
        text_type: TextTypes = TextTypes.PRINTED,
        confidence: float = 0,
        is_clickable: bool = False,
        is_structure: bool = False,
    ):
        super().__init__(entity_id, bbox)
        self._text = text
        self._text_type = text_type
        self._confidence = confidence / 100
        self.is_clickable = is_clickable
        self.is_structure = is_structure
        self._page = None
        self._page_id = None
        self.line = None
        self.cell_bbox = None
        self.cell_id = None
        self.row_index = None
        self.col_index = None
        self.row_span = None
        self.col_span = None
        self.key_bbox = None
        self.value_bbox = None
        self.key_id = None
        self.value_id = None
        self.kv_id = None
        self.kv_bbox = None
        self.line_id = None
        self.line_bbox = None
        self.table_id = None
        self.table_bbox = None
        self.layout_id = None
        self.layout_type = None
        self.layout_bbox = None

    @property
    def text(self) -> str:
        """
        :return: Returns the text transcription of the Word entity.
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, text: str):
        """
        Sets the transcription of the :class:`Word` entity.

        :param text: String containing the text transcription of the Word entity.
        :type text: str
        """
        self._text = text

    @property
    def text_type(self) -> TextTypes:
        """
        :return: Returns the property of :class:`Word` class that holds the text type of Word object.
        :rtype: str
        """
        return self._text_type

    @text_type.setter
    def text_type(self, text_type: TextTypes):
        """
        Sets the text type of the :class:`Word` entity.

        :param text_type: Constant containing the type of text
        """
        assert isinstance(text_type, TextTypes)
        self._text_type = text_type

    @property
    def page(self) -> int:
        """
        :return: Returns the page number of the page the Word entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`Word` entity.

        :param page_num: Page number where the Word entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the Word entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    @property
    def words(self):
        """Returns itself

        :rtype: Word
        """
        return [self]

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        return self.text, [self]

    def __repr__(self) -> str:
        """
        :return: String representation of the Word entity.
        :rtype: str
        """
        return f"{self.text}"



import os
import sys
import subprocess
import setuptools
from setuptools import find_packages, setup
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def read_requirements(path):
    with open(path, "r") as f:
        requirements = [line for line in f.readlines()]
    return requirements

class TestCommand(setuptools.Command):

    description = 'run linters, tests and create a coverage report'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        #self._run(['pytest', 'tests/'])
        return

    def _run(self, command):
        try:
            subprocess.check_call(command)
        except subprocess.CalledProcessError as error:
            print('Command failed with exit code', error.returncode)
            sys.exit(error.returncode)

setup(
    # include data files
    name="amazon-textract-textractor",
    version="1.8.1",
    license="Apache 2.0",
    description="A package to use AWS Textract services.",
    url="https://github.com/aws-samples/amazon-textract-textractor",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="amazon textract aws ocr document",
    packages=find_packages(exclude=["docs", "tests"], ),
    include_package_data=True,
    install_requires=read_requirements(os.path.join(here, "requirements.txt")),
    extras_require={
        f.split(".")[0]: read_requirements(os.path.join(here, "extras", f))
        for f in os.listdir(os.path.join(here, "extras"))
    },
    cmdclass={'test': TestCommand},
    test_command="test",
    entry_points={
        "console_scripts": [
            "textractor = textractor.cli.cli:textractor_cli",
        ],
    },
)



"""
Represents a single :class:`Value` Entity within the :class:`Document`. 
The Textract API response returns groups of words as KEY_VALUE_SET BlockTypes. These may be of KEY
or VALUE type which is indicated by the EntityType attribute in the JSON response. 

This class contains the associated metadata with the :class:`Value` entity including the entity ID, 
bounding box information, child words, associated key ID, page number, Page ID, confidence of detection
and if it's a checkbox.
"""

from typing import List
import uuid

from textractor.entities.word import Word
from textractor.exceptions import InputError
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.data.constants import PRINTED, HANDWRITING, TextTypes
from textractor.visualizers.entitylist import EntityList
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.utils.text_utils import linearize_children


class Value(DocumentEntity):
    """
    To create a new :class:`Value` object we need the following:

    :param entity_id: Unique identifier of the Word entity.
    :type entity_id: str
    :param bbox: Bounding box of the Word entity.
    :type bbox: BoundingBox
    :param confidence: value storing the confidence of detection out of 100.
    :type confidence: float
    """

    def __init__(self, entity_id: str, bbox: BoundingBox, confidence: float = 0):
        super().__init__(entity_id, bbox)
        self._words: List[Word] = []
        self._key_id = None
        self._contains_checkbox = False
        self._confidence = confidence / 100
        self._page = None
        self._page_id = None

    @property
    def words(self) -> List[Word]:
        """
        :return: Returns a list of all words in the entity if it exists else returns the
                 checkbox status of the Value entity.
        :rtype: EntityList[Word]
        """
        if not self.contains_checkbox:
            return EntityList(self._words)
        else:
            return self.children[0].words

    @words.setter
    def words(self, words: List[Word]):
        """
        Add Word objects to the Value.

        :param words: List of Word objects, each representing a word within the Value field.
        No specific ordering is assumed as is it is ordered internally.
        :type words: list
        """
        self._words = sorted(words, key=lambda x: x.bbox.x + x.bbox.y)

    @property
    def key_id(self) -> str:
        """
        Returns the associated Key ID for the :class:`Value` entity.

        :return: Returns the associated KeyValue object ID.
        :rtype: str
        """
        return self._key_id

    @key_id.setter
    def key_id(self, key_id: str):
        """
        Sets the :class:`KeyValue` ID to which this Value object is associated.

        :param key_id: Unique identifier for the KeyValue entity to which this Value is associated.
        :type key_id: str
        """
        self._key_id = key_id

    @property
    def contains_checkbox(self) -> bool:
        """
        Returns True if the value associated is a :class:`SelectionElement`.

        :return: Returns True if the value associated is a checkbox/SelectionElement.
        :rtype: bool
        """
        return self._contains_checkbox

    @contains_checkbox.setter
    def contains_checkbox(self, checkbox_present: bool):
        """
        Sets the contains_checkbox field to True/False depending on the presence of a :class:`SelectionElement`.

        :param checkbox_present: True or False depending on checkbox presence.
        :type checkbox_present: bool
        """
        self._contains_checkbox = checkbox_present

    @property
    def page(self):
        """
        :return: Returns the page number of the page the Value entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the Value entity.

        :param page_num: Page number where the Value entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Value` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    def get_words_by_type(self, text_type: str = TextTypes.PRINTED) -> List[Word]:
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes

        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        return EntityList([word for word in self.words if word.text_type == text_type])

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        # We will need to handle the case where the Value contains a checkbox.
        if self.contains_checkbox:
            text, words = self.children[0].get_text_and_words(config)
        else:
            text, words = linearize_children(
                self.words,
                config=config,
                no_new_lines=config.remove_new_lines_in_leaf_elements,
            )
        if config.add_prefixes_and_suffixes_in_text:
            text = config.value_prefix + text + config.value_suffix
        if config.add_prefixes_and_suffixes_as_words:
            words = (
                (
                    [
                        Word(
                            str(uuid.uuid4()),
                            self.bbox,
                            config.value_prefix,
                            is_structure=True,
                            is_clickable=(
                                bool(words) and words[0] in [config.selection_element_selected, config.selection_element_not_selected]
                            )
                        )
                    ]
                    if config.value_prefix else
                    []
                ) +
                words +
                (
                    [
                        Word(
                            str(uuid.uuid4()),
                            self.bbox,
                            config.value_suffix,
                            is_structure=True,
                            is_clickable=(
                                bool(words) and words[0] in [config.selection_element_selected, config.selection_element_not_selected]
                            )
                        )
                    ]
                    if config.value_suffix else
                    []
                )
            )
        for w in words:
            w.value_id = str(self.id)
            w.value_bbox = self.bbox
        return text, words

    def __repr__(self) -> str:
        """
        :return: String representation of the Value entity.
        :rtype: str
        """
        repr_string = ""
        if self.contains_checkbox:
            repr_string += f"{self.children[0].__repr__()}"
        else:
            words = " ".join([word.text for word in self.words])
            repr_string += f"{words}"
        return repr_string



""":class:`DocumentEntity` is the class that all Document entities such as :class:`Word`, :class:`Line`, :class:`Table` etc. inherit from. This class provides methods 
useful to all such entities."""

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple
from textractor.entities.bbox import BoundingBox
from textractor.visualizers.entitylist import EntityList
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.data.html_linearization_config import HTMLLinearizationConfig
from textractor.data.markdown_linearization_config import MarkdownLinearizationConfig
from textractor.entities.linearizable import Linearizable

class DocumentEntity(Linearizable, ABC):
    """
    An interface for all document entities within the document body, composing the
    hierarchy of the document object model.
    The purpose of this class is to define properties common to all document entities
    i.e. unique id and bounding box.
    """

    def __init__(self, entity_id: str, bbox: BoundingBox):
        """
        Initialize the common properties to DocumentEntities. Additionally, it contains information about
        child entities within a document entity.

        :param entity_id: Unique identifier of the document entity.
        :param bbox: Bounding box of the entity
        """
        self.id = entity_id
        self._bbox: BoundingBox = bbox
        self.metadata = {}  # Holds optional information about the entity
        self._children = []
        self._children_type = None
        self._raw_object = None

    def add_children(self, children):
        """
        Adds children to all entities that have parent-child relationships.

        :param children: List of child entities.
        :type children: list
        """
        self._children.extend(children)

    @property
    def raw_object(self) -> Dict:
        """
        :return: Returns the raw dictionary object that was used to create this Python object
        :rtype: Dict
        """
        return self._raw_object

    @raw_object.setter
    def raw_object(self, raw_object: Dict):
        """
        Sets the raw object that was used to create this Python object
        :param raw_object: raw object dictionary from the response
        :type raw_object: Dict
        """
        self._raw_object = raw_object

    @property
    def x(self) -> float:
        """
        :return: Returns x coordinate for bounding box
        :rtype: float
        """
        return self._bbox.x

    @x.setter
    def x(self, x: float):
        """
        Sets x coordinate for bounding box

        :param x: x coordinate of the bounding box
        :type x: float
        """
        self._bbox.x = x

    @property
    def y(self) -> float:
        """
        :return: Returns y coordinate for bounding box
        :rtype: float
        """
        return self._bbox.y

    @y.setter
    def y(self, y: float):
        """
        Sets y coordinate for bounding box.

        :param y: y coordinate of the bounding box
        :type y: float
        """
        self._bbox.y = y

    @property
    def width(self) -> float:
        """
        :return: Returns width for bounding box
        :rtype: float
        """
        return self._bbox.width

    @width.setter
    def width(self, width: float):
        """
        Sets width for bounding box.

        :param width: width of the bounding box
        :type width: float
        """
        self._bbox.width = width

    @property
    def height(self) -> float:
        """
        :return: Returns height for bounding box
        :rtype: float
        """
        return self._bbox.height

    @height.setter
    def height(self, height: float):
        """
        Sets height for bounding box.

        :param height: height of the bounding box
        :type height: float
        """
        self._bbox.height = height

    @property
    def bbox(self) -> BoundingBox:
        """
        :return: Returns entire bounding box of entity
        :rtype: BoundingBox
        """
        return self._bbox

    @bbox.setter
    def bbox(self, bbox: BoundingBox):
        self._bbox = bbox

    @property
    def children(self):
        """
        :return: Returns children of entity
        :rtype: list
        """
        return self._children

    @property
    def confidence(self) -> float:
        """
        Returns the object confidence as predicted by Textract. If the confidence is not available, returns None

        :return: Prediction confidence for a document entity, between 0 and 1
        :rtype: float
        """
        
        # Needed for backward compatibility
        if hasattr(self, "_confidence"):
            return self._confidence
    
        # Uses the raw response
        if not hasattr(self, "raw_object"):
            return None
        confidence = self.raw_object.get("Confidence")
        if confidence is None:
            return None            
        return confidence / 100

    def remove(self, entity):
        """
        Recursively removes an entity from the child tree of a document entity and update its bounding box

        :param entity: Entity
        :type entity: DocumentEntity
        """

        for c in self._children:
            if entity == c:
                # Element was found in children
                break
            if not c.__class__.__name__ == "Word" and c.remove(entity):
                # Element was found and removed in grand-children
                if not c.children:
                    self._children.remove(c)
                return True
        else:
            # Element was not found
            return False
        self._children.remove(c)
        if self._children:
            self.bbox = BoundingBox.enclosing_bbox(self._children)
        return True

    def visit(self, word_set):
        for c in list(self._children):
            if c.__class__.__name__ == "Word":
                if c.id in word_set:
                    self._children.remove(c)
                else:
                    word_set.add(c.id)
            else:
                c.visit(word_set)

    def visualize(self, *args, **kwargs) -> EntityList:
        """
        Returns the object's children in a visualization EntityList object

        :return: Returns an EntityList object
        :rtype: EntityList
        """
        return EntityList(self).visualize(*args, **kwargs)



from typing import Union, Tuple
from io import BytesIO
from PIL import Image

from textractor.exceptions import InputError


def s3_path_to_bucket_and_prefix(s3_path: str) -> Tuple[str, str]:
    """Converts an S3 URI to a bucket and prefix

    :param s3_path: S3 URI
    :type s3_path: str
    :raises InputError: Raised if the given path cannot be parsed
    :return: Tuple of bucket and prefix as string
    :rtype: Tuple[str, str]
    """
    try:
        bucket, prefix = s3_path.replace("s3://", "").split("/", 1)
    except IndexError:
        raise InputError(f"Could not parse {s3_path} as ")
    return bucket, prefix


def download_from_s3(client, s3_path: str, **extra_args):
    """Downloads a file from S3 and returns it as a BytesIO object

    :param client: S3 client
    :type client: Client
    :param s3_path: S3 path to download
    :type s3_path: str
    :return: S3 file as a BytesIO object
    :rtype: BytesIO
    """

    bucket, prefix = s3_path_to_bucket_and_prefix(s3_path)

    f = BytesIO()
    client.download_fileobj(bucket, prefix, f)
    f.seek(0)
    return f


def upload_to_s3(
    client, s3_path: str, file_source: Union[str, bytes, Image.Image], **extra_args
):
    """Upload a file to S3

    :param client: boto3 client
    :type client: Client
    :param s3_path: S3 path to upload to
    :type s3_path: str
    :param file_source: File to upload
    :type file_source: Union[str, bytes, Image.Image]
    :raises InputError: Raised if the file_source is not of type str, bytes or Image
    """
    bucket, prefix = s3_path_to_bucket_and_prefix(s3_path)
    if isinstance(file_source, Image.Image):
        fake_file = BytesIO()
        file_source.save(fake_file, format="PNG")
        fake_file.seek(0)
        client.upload_fileobj(fake_file, bucket, prefix, extra_args)
    elif isinstance(file_source, bytes):
        fake_file = BytesIO(file_source)
        client.upload_fileobj(fake_file, bucket, prefix, extra_args)
    elif isinstance(file_source, str):
        client.upload_file(file_source, bucket, prefix, extra_args)
    else:
        raise InputError(
            f"{file_source} must be of type str or bytes, not {type(file_source)}"
        )


def delete_from_s3(client, s3_path: str):
    """Delete a file from S3

    :param client: boto3 client
    :type client: Client
    :param s3_path: S3 path to the object to delete
    :type s3_path: str
    """
    bucket, prefix = s3_path_to_bucket_and_prefix(s3_path)

    client.delete_object(bucket, prefix)
